// The Exchange: Genetically Evolving Market Agents 
// Written by David J. Stein, Esq., 11/22/2004. See "License.html" for terms and conditions of use.
//
// This module contains all of the code for an Exchange Agent and the Exchange Virtual Machine in which the agents execute.

#region Using declarations

using System;
using System.Web.Mail;
using System.Xml;
using System.Xml.Schema;
using System.Windows.Forms;
using System.Collections;

#endregion

namespace The_Exchange_Namespace {

	public abstract class ExchangeAgent {
		// This abstract class defines common operations for an Exchange Agent, which controls a player in the exchange simulation.

		#region Variables and initialization code

		public Player player;
		public string strName;
		public int iBuyGood;
		public int iSellGood;
		public int iProduction;
		public int iGames;
		public double dRank;
		public long lValue;
		public double dFitness;
		public ExchangeSimulation exchange;

		public ExchangeAgent()	{
			strName = "Empty Agent";
			iBuyGood = iSellGood = -1;
			iProduction = 0;
			iGames = 0;
			dFitness = 0.1f;
			dRank = 0.1f;
			lValue = 0;
		}

		#endregion

		#region Core functions

		public abstract void Process();  // This function is called by the simulation framework once per second. It's overridden in each subclass of ExchangeAgent.

		#endregion

	}

	public class ExchangeAgent_Console : ExchangeAgent {
		// This class represents the console player.

		#region Variables and initialization code

		public ExchangeAgent_Console(string strName) : base() {
			this.strName = strName;
		}

		#endregion

		#region Core functions

		public override void Process() {  // do nothing - no AI function here.
		}

		#endregion
	}

	public class ExchangeAgent_Programmed : ExchangeAgent {
		// This class represents a programmed AI player.

		#region Variables and initialization code

		public const int iProgramSize = 256;						// defines the size of the program.
		public const int iProgramTimeout = 4096;			// defines the maximum number of instructions to execute before timing out (to break infinite loops.)
		public char[] cProgram;														// the program memory space.
		public string strAssembly = "";										// the assembly code (if available) - only for display purposes.
		private ExchangeVirtualMachine machine;			// the virtual machine in which this agent executes
		private static string XmlLoadErrors;

		public ExchangeAgent_Programmed() : base() {
			machine = new ExchangeVirtualMachine();
			// create random program - note: for all randomly-generated entities, prevent all backward jumps, since they have a great tendency to produce infinite loops.
			cProgram = new char[ExchangeAgent_Programmed.iProgramSize];
		}

		public ExchangeAgent_Programmed(Random r) : base() {
			strName = "A.I. " + r.Next(1000000).ToString();
			machine = new ExchangeVirtualMachine();
			// create random program - note: for all randomly-generated entities, prevent all backward jumps, since they have a great tendency to produce infinite loops.
			cProgram = new char[ExchangeAgent_Programmed.iProgramSize];
			bool bLastJump = false;
			for (int i = 0; i < ExchangeAgent_Programmed.iProgramSize; i++) {
				cProgram[i] = (char) r.Next(256);
				if (bLastJump == true) {
					cProgram[i] = (char) ((int) cProgram[i] & 127);  // remove high-order bit so that this signed char is never negative.
					bLastJump = false;
				}
				else
					bLastJump = ((cProgram[i] >= 138) && (cProgram[i] <= 146));
			}
		}

		#endregion

		#region Core functions

		public override void Process() {
			machine.Execute(this);
		}

		public ArrayList DisassembleProgram() {
			// This function disassembles the machine opcodes and returns a list of assembly instructions.
			ArrayList list = new ArrayList();
			int iProgramCounter = 0;
			while (iProgramCounter < ExchangeAgent_Programmed.iProgramSize) {
				char cInstruction = cProgram[iProgramCounter];
				iProgramCounter++;
				// parse opcode
				if (cInstruction == 0)
					list.Add(((int) (iProgramCounter - 1)).ToString("D3") + "\t" + ((int) cInstruction).ToString("X2") + "\t" + "NOOP");
				else if ((cInstruction >= 1) && (cInstruction <= 4))
					list.Add(((int) (iProgramCounter - 1)).ToString("D3") + "\t" + ((int) cInstruction).ToString("X2") + "\t" + "CLR R" + (cInstruction - 1));
				else if ((cInstruction >= 5) && (cInstruction <= 8) && (iProgramCounter < cProgram.Length)) {
					list.Add(((int) (iProgramCounter - 1)).ToString("D3") + '\t' + ((int) cInstruction).ToString("X2") + " " + ((int) cProgram[iProgramCounter]).ToString("X2") + "\t" + "SET R" + (cInstruction - 5) + ", " + (cProgram[iProgramCounter] > 128 ? cProgram[iProgramCounter] : (int) cProgram[iProgramCounter]));
					iProgramCounter++;
				}
				else if ((cInstruction >= 9) && (cInstruction <= 20)) {
					int iDestRegister = (cInstruction - 9) / 3;
					int iSrcRegister = (cInstruction - 9) % 3;
					if (iSrcRegister >= iDestRegister)
						iSrcRegister++;
					list.Add(((int) (iProgramCounter - 1)).ToString("D3") + "\t" + ((int) cInstruction).ToString("X2") + "\t" + "SET R" + iDestRegister + ", R" + iSrcRegister);
				}
				else if ((cInstruction >= 21) && (cInstruction <= 32)) {
					int iDestRegister = (cInstruction - 21) / 3;
					int iSrcRegister = (cInstruction - 21) % 3;
					if (iSrcRegister >= iDestRegister)
						iSrcRegister++;
					list.Add(((int) (iProgramCounter - 1)).ToString("D3") + "\t" + ((int) cInstruction).ToString("X2") + "\t" + "SWAP R" + iDestRegister + ", R" + iSrcRegister);
				}
				else if ((cInstruction >= 33) && (cInstruction <= 41))
					list.Add(((int) (iProgramCounter - 1)).ToString("D3") + "\t" + ((int) cInstruction).ToString("X2") + "\t" + "LD R0, PRICE" + (cInstruction - 32));
				else if ((cInstruction >= 42) && (cInstruction <= 50))
					list.Add(((int) (iProgramCounter - 1)).ToString("D3") + "\t" + ((int) cInstruction).ToString("X2") + "\t" + "LD R0, TREND" + (cInstruction - 41));
				else if ((cInstruction >= 51) && (cInstruction <= 59))
					list.Add(((int) (iProgramCounter - 1)).ToString("D3") + "\t" + ((int) cInstruction).ToString("X2") + "\t" + "LD R0, QUANTITY" + (cInstruction - 50));
				else if ((cInstruction >= 60) && (cInstruction <= 68))
					list.Add(((int) (iProgramCounter - 1)).ToString("D3") + "\t" + ((int) cInstruction).ToString("X2") + "\t" + "LD R0, HOLDINGS" + (cInstruction - 59));
				else if (cInstruction == 69)
					list.Add(((int) (iProgramCounter - 1)).ToString("D3") + "\t" + ((int) cInstruction).ToString("X2") + "\t" + "LD R0, CREDITS");
				else if (cInstruction == 70)
					list.Add(((int) (iProgramCounter - 1)).ToString("D3") + "\t" + ((int) cInstruction).ToString("X2") + "\t" + "LD R0, EVENT");
				else if (cInstruction == 71)
					list.Add(((int) (iProgramCounter - 1)).ToString("D3") + "\t" + ((int) cInstruction).ToString("X2") + "\t" + "LD R0, PHASE");
				else if (cInstruction == 72)
					list.Add(((int) (iProgramCounter - 1)).ToString("D3") + "\t" + ((int) cInstruction).ToString("X2") + "\t" + "LD R0, PLAYERCOUNT");
				else if (cInstruction == 73)
					list.Add(((int) (iProgramCounter - 1)).ToString("D3") + "\t" + ((int) cInstruction).ToString("X2") + "\t" + "LD R0, RANDOM");
				else if ((cInstruction >= 74) && (cInstruction <= 77) && (iProgramCounter < cProgram.Length)) {
					list.Add(((int) (iProgramCounter - 1)).ToString("D3") + '\t' + ((int) cInstruction).ToString("X2") + " " + ((int) cProgram[iProgramCounter]).ToString("X2") + "\t" + "ADD R" + (cInstruction - 74) + ", " + (int) cProgram[iProgramCounter]);
					iProgramCounter++;
				}
				else if ((cInstruction >= 78) && (cInstruction <= 89)) {
					int iDestRegister = (cInstruction - 78) / 3;
					int iSrcRegister = (cInstruction - 78) % 3;
					if (iSrcRegister >= iDestRegister)
						iSrcRegister++;
					list.Add(((int) (iProgramCounter - 1)).ToString("D3") + "\t" + ((int) cInstruction).ToString("X2") + "\t" + "ADD R" + iDestRegister + ", R" + iSrcRegister);
				}
				else if ((cInstruction >= 90) && (cInstruction <= 93) && (iProgramCounter < cProgram.Length)) {
					list.Add(((int) (iProgramCounter - 1)).ToString("D3") + '\t' + ((int) cInstruction).ToString("X2") + " " + ((int) cProgram[iProgramCounter]).ToString("X2") + "\t" + "SUB R" + (cInstruction - 90) + ", " + (int) cProgram[iProgramCounter]);
					iProgramCounter++;
				}
				else if ((cInstruction >= 94) && (cInstruction <= 105)) {
					int iDestRegister = (cInstruction - 94) / 3;
					int iSrcRegister = (cInstruction - 94) % 3;
					if (iSrcRegister >= iDestRegister)
						iSrcRegister++;
					list.Add(((int) (iProgramCounter - 1)).ToString("D3") + "\t" + ((int) cInstruction).ToString("X2") + "\t" + "SUB R" + iDestRegister + ", R" + iSrcRegister);
				}
				else if ((cInstruction >= 106) && (cInstruction <= 109) && (iProgramCounter < cProgram.Length)) {
					list.Add(((int) (iProgramCounter - 1)).ToString("D3") + '\t' + ((int) cInstruction).ToString("X2") + " " + ((int) cProgram[iProgramCounter]).ToString("X2") + "\t" + "MLT R" + (cInstruction - 106) + ", " + (int) cProgram[iProgramCounter]);
					iProgramCounter++;
				}
				else if ((cInstruction >= 110) && (cInstruction <= 121)) {
					int iDestRegister = (cInstruction - 110) / 3;
					int iSrcRegister = (cInstruction - 110) % 3;
					if (iSrcRegister >= iDestRegister)
						iSrcRegister++;
					list.Add(((int) (iProgramCounter - 1)).ToString("D3") + "\t" + ((int) cInstruction).ToString("X2") + "\t" + "MLT R" + iDestRegister + ", R" + iSrcRegister);
				}
				else if ((cInstruction >= 122) && (cInstruction <= 125) && (iProgramCounter < cProgram.Length)) {
					list.Add(((int) (iProgramCounter - 1)).ToString("D3") + '\t' + ((int) cInstruction).ToString("X2") + " " + ((int) cProgram[iProgramCounter]).ToString("X2") + "\t" + "DIV R" + (cInstruction - 122) + ", " + (int) cProgram[iProgramCounter]);
					iProgramCounter++;
				}
				else if ((cInstruction >= 126) && (cInstruction <= 137)) {
					int iDestRegister = (cInstruction - 126) / 3;
					int iSrcRegister = (cInstruction - 126) % 3;
					if (iSrcRegister >= iDestRegister)
						iSrcRegister++;
					list.Add(((int) (iProgramCounter - 1)).ToString("D3") + "\t" + ((int) cInstruction).ToString("X2") + "\t" + "DIV R" + iDestRegister + ", R" + iSrcRegister);
				}
				else if ((cInstruction == 138) && (iProgramCounter < cProgram.Length)) {
					list.Add(((int) (iProgramCounter - 1)).ToString("D3") + "\t" + ((int) cInstruction).ToString("X2") + " " + ((int) cProgram[iProgramCounter]).ToString("X2") + "\t" + "JMP " + (cProgram[iProgramCounter] > 128 ? cProgram[iProgramCounter] - 256 : (int) cProgram[iProgramCounter]));
					iProgramCounter++;
				}
				else if ((cInstruction >= 139) && (cInstruction <= 142) && (iProgramCounter < cProgram.Length)) {
					list.Add(((int) (iProgramCounter - 1)).ToString("D3") + '\t' + ((int) cInstruction).ToString("X2") + " " + ((int) cProgram[iProgramCounter]).ToString("X2") + "\t" + "JZ R" + (cInstruction - 139) + ", " + (cProgram[iProgramCounter] > 128 ? cProgram[iProgramCounter] - 256 : (int) cProgram[iProgramCounter]));
					iProgramCounter++;
				}
				else if ((cInstruction >= 143) && (cInstruction <= 146) && (iProgramCounter < cProgram.Length)) {
					list.Add(((int) (iProgramCounter - 1)).ToString("D3") + '\t' + ((int) cInstruction).ToString("X2") + " " + ((int) cProgram[iProgramCounter]).ToString("X2") + "\t" + "JNZ R" + (cInstruction - 143) + ", " + (cProgram[iProgramCounter] > 128 ? cProgram[iProgramCounter] - 256 : (int) cProgram[iProgramCounter]));
					iProgramCounter++;
				}
				else if ((cInstruction >= 147) && (cInstruction <= 150) && (iProgramCounter < cProgram.Length)) {
					list.Add(((int) (iProgramCounter - 1)).ToString("D3") + '\t' + ((int) cInstruction).ToString("X2") + " " + ((int) cProgram[iProgramCounter]).ToString("X2") + "\t" + "CMPE R" + (cInstruction - 147) + ", " + (int) cProgram[iProgramCounter]);
					iProgramCounter++;
				}
				else if ((cInstruction >= 151) && (cInstruction <= 162)) {
					int iDestRegister = (cInstruction - 151) / 3;
					int iSrcRegister = (cInstruction - 151) % 3;
					if (iSrcRegister >= iDestRegister)
						iSrcRegister++;
					list.Add(((int) (iProgramCounter - 1)).ToString("D3") + "\t" + ((int) cInstruction).ToString("X2") + "\t" + "CMPE R" + iDestRegister + ", R" + iSrcRegister);
				}
				else if ((cInstruction >= 163) && (cInstruction <= 166) && (iProgramCounter < cProgram.Length)) {
					list.Add(((int) (iProgramCounter - 1)).ToString("D3") + '\t' + ((int) cInstruction).ToString("X2") + " " + ((int) cProgram[iProgramCounter]).ToString("X2") + "\t" + "CMPNE R" + (cInstruction - 163) + ", " + (int) cProgram[iProgramCounter]);
					iProgramCounter++;
				}
				else if ((cInstruction >= 167) && (cInstruction <= 178)) {
					int iDestRegister = (cInstruction - 167) / 3;
					int iSrcRegister = (cInstruction - 167) % 3;
					if (iSrcRegister >= iDestRegister)
						iSrcRegister++;
					list.Add(((int) (iProgramCounter - 1)).ToString("D3") + "\t" + ((int) cInstruction).ToString("X2") + "\t" + "CMPNE R" + iDestRegister + ", R" + iSrcRegister);
				}
				else if ((cInstruction >= 179) && (cInstruction <= 182) && (iProgramCounter < cProgram.Length)) {
					list.Add(((int) (iProgramCounter - 1)).ToString("D3") + '\t' + ((int) cInstruction).ToString("X2") + " " + ((int) cProgram[iProgramCounter]).ToString("X2") + "\t" + "CMPG R" + (cInstruction - 179) + ", " + (int) cProgram[iProgramCounter]);
					iProgramCounter++;
				}
				else if ((cInstruction >= 183) && (cInstruction <= 194)) {
					int iDestRegister = (cInstruction - 183) / 3;
					int iSrcRegister = (cInstruction - 183) % 3;
					if (iSrcRegister >= iDestRegister)
						iSrcRegister++;
					list.Add(((int) (iProgramCounter - 1)).ToString("D3") + "\t" + ((int) cInstruction).ToString("X2") + "\t" + "CMPG R" + iDestRegister + ", R" + iSrcRegister);
				}
				else if ((cInstruction >= 195) && (cInstruction <= 198) && (iProgramCounter < cProgram.Length)) {
					list.Add(((int) (iProgramCounter - 1)).ToString("D3") + '\t' + ((int) cInstruction).ToString("X2") + " " + ((int) cProgram[iProgramCounter]).ToString("X2") + "\t" + "CMPGE R" + (cInstruction - 195) + ", " + (int) cProgram[iProgramCounter]);
					iProgramCounter++;
				}
				else if ((cInstruction >= 199) && (cInstruction <= 210)) {
					int iDestRegister = (cInstruction - 199) / 3;
					int iSrcRegister = (cInstruction - 199) % 3;
					if (iSrcRegister >= iDestRegister)
						iSrcRegister++;
					list.Add(((int) (iProgramCounter - 1)).ToString("D3") + "\t" + ((int) cInstruction).ToString("X2") + "\t" + "CMPGE R" + iDestRegister + ", R" + iSrcRegister);
				}
				else if ((cInstruction >= 211) && (cInstruction <= 214) && (iProgramCounter < cProgram.Length)) {
					list.Add(((int) (iProgramCounter - 1)).ToString("D3") + '\t' + ((int) cInstruction).ToString("X2") + " " + ((int) cProgram[iProgramCounter]).ToString("X2") + "\t" + "CMPL R" + (cInstruction - 211) + ", " + (int) cProgram[iProgramCounter]);
					iProgramCounter++;
				}
				else if ((cInstruction >= 215) && (cInstruction <= 226)) {
					int iDestRegister = (cInstruction - 215) / 3;
					int iSrcRegister = (cInstruction - 215) % 3;
					if (iSrcRegister >= iDestRegister)
						iSrcRegister++;
					list.Add(((int) (iProgramCounter - 1)).ToString("D3") + "\t" + ((int) cInstruction).ToString("X2") + "\t" + "CMPL R" + iDestRegister + ", R" + iSrcRegister);
				}
				else if ((cInstruction >= 227) && (cInstruction <= 230) && (iProgramCounter < cProgram.Length)) {
					list.Add(((int) (iProgramCounter - 1)).ToString("D3") + '\t' + ((int) cInstruction).ToString("X2") + " " + ((int) cProgram[iProgramCounter]).ToString("X2") + "\t" + "CMPLE R" + (cInstruction - 227) + ", " + (int) cProgram[iProgramCounter]);
					iProgramCounter++;
				}
				else if ((cInstruction >= 231) && (cInstruction <= 242)) {
					int iDestRegister = (cInstruction - 231) / 3;
					int iSrcRegister = (cInstruction - 231) % 3;
					if (iSrcRegister >= iDestRegister)
						iSrcRegister++;
					list.Add(((int) (iProgramCounter - 1)).ToString("D3") + "\t" + ((int) cInstruction).ToString("X2") + "\t" + "CMPLE R" + iDestRegister + ", R" + iSrcRegister);
				}
				else if ((cInstruction >= 243) && (cInstruction <= 246))
					list.Add(((int) (iProgramCounter - 1)).ToString("D3") + "\t" + ((int) cInstruction).ToString("X2") + "\t" + "SET BUY, R" + (cInstruction - 243));
				else if ((cInstruction >= 247) && (cInstruction <= 250))
					list.Add(((int) (iProgramCounter - 1)).ToString("D3") + "\t" + ((int) cInstruction).ToString("X2") + "\t" + "SET SELL, R" + (cInstruction - 247));
				else if ((cInstruction >= 251) && (cInstruction <= 254))
					list.Add(((int) (iProgramCounter - 1)).ToString("D3") + "\t" + ((int) cInstruction).ToString("X2") + "\t" + "SET PRODUCE, R" + (cInstruction - 251));
				else if (cInstruction == 255)
					list.Add(((int) (iProgramCounter - 1)).ToString("D3") + "\t" + ((int) cInstruction).ToString("X2") + "\t" + "END");
				else
					list.Add(((int) (iProgramCounter - 1)).ToString("D3") + "\t" + ((int) cInstruction).ToString("X2") + "\t" + "Unknown Opcode " + cInstruction);
			}
			// finish and return list of disassembled instructions
			return list;
		}

		public void Save(string filename) {
			// This function saves to disk a programmed ExchangeAgent. Implicitly saves according to the ExchangeAgent.xsd definition, but does not rely on that file.
			XmlTextWriter save = new XmlTextWriter(filename, System.Text.Encoding.UTF8);
			save.Formatting = Formatting.Indented;
			save.IndentChar = ' ';
			save.WriteStartDocument();
			save.WriteStartElement("ExchangeAgent");
			save.Indentation++;
			// write parameters
			save.WriteStartElement("Parameters");
			save.Indentation++;
			save.WriteElementString("Name", strName);
			save.WriteElementString("GamesPlayed", iGames.ToString());
			save.WriteElementString("Fitness", dFitness.ToString());
			save.Indentation--;
			save.WriteEndElement();
			// write program
			save.WriteStartElement("Program");
			save.Indentation++;
			// write assembly, if provided
			if (strAssembly.Length > 0)
				save.WriteElementString("Assembly", strAssembly);
			for (int i = 0; i < ExchangeAgent_Programmed.iProgramSize; i++) {
				int opcode = (int) cProgram[i];
				save.WriteElementString("Opcode", opcode.ToString());
			}
			save.Indentation--;
			save.WriteEndElement();
			// close XML file
			save.Indentation--;
			save.WriteEndElement();
			save.WriteEndDocument();
			save.Close();
		}

		public static ExchangeAgent_Programmed Load(string file) {
			// This function loads an ExchangeAgent from disk. This file explicitly relies on the ExchangeAgent.xsd definition.
			try {
				ExchangeAgent_Programmed agent = new ExchangeAgent_Programmed(new Random(0));
				XmlTextReader xtr = new XmlTextReader(file);
				XmlValidatingReader xvr = new XmlValidatingReader(xtr);
				xvr.ValidationType = ValidationType.Schema;
				xvr.ValidationEventHandler += new ValidationEventHandler(ValidationErrors);
				XmlSchemaCollection xsc = new XmlSchemaCollection();
				xsc.Add("", System.Reflection.Assembly.GetExecutingAssembly().Location.Substring(0, System.Reflection.Assembly.GetExecutingAssembly().Location.Length - 16) + @"4 Resources\ExchangeAgent.xsd");
				xvr.Schemas.Add(xsc);
				XmlDocument xd = new XmlDocument();
				XmlLoadErrors = "";
				xd.Load(xvr);
				if (XmlLoadErrors.Length > 0)
					MessageBox.Show("Errors loading network document. This does not appear to be a valid Exchange agent.", "Errors Loading Exchange Agent", MessageBoxButtons.OK, MessageBoxIcon.Error);
				else {
					foreach (XmlNode x1 in xd.ChildNodes) {
						if (x1.Name == "ExchangeAgent") {
							foreach (XmlNode x2 in x1.ChildNodes) {
								if (x2.Name == "Parameters") {
									foreach (XmlNode x3 in x2.ChildNodes) {
										if (x3.Name == "Name")
											agent.strName = x3.InnerText;
										else if (x3.Name == "GamesPlayed")
											agent.iGames = Int32.Parse(x3.InnerText);
										else if (x3.Name == "Fitness")
											agent.dFitness = double.Parse(x3.InnerText);
									}
								}
								else if (x2.Name == "Program") {
									int count = 0;
									foreach (XmlNode x3 in x2.ChildNodes) {
										if (x3.Name == "Assembly")
											agent.strAssembly = x3.InnerText;
										else if (x3.Name == "Opcode")
											agent.cProgram[count++] = (char) Int32.Parse(x3.InnerText);
									}
								}
							}
						}
					}
					xtr.Close();
					return agent;
				}
			} catch (Exception ex) {
				MessageBox.Show(ex.Message);
				return null;
			}
			return null;
		}

		public static void SaveSet(string filename, ArrayList agents) {
			// This function saves to disk an entire set of programmed ExchangeAgents. Implicitly saves according to the ExchangeAgentSet.xsd definition, but does not rely on that file.
			XmlTextWriter save = new XmlTextWriter(filename, System.Text.Encoding.UTF8);
			save.Formatting = Formatting.Indented;
			save.IndentChar = ' ';
			save.WriteStartDocument();
			save.WriteStartElement("ExchangeAgentSet");
			save.Indentation++;
			foreach (ExchangeAgent_Programmed agent in agents) {
				save.WriteStartElement("ExchangeAgent");
				save.Indentation++;
				// write parameters
				save.WriteStartElement("Parameters");
				save.Indentation++;
				save.WriteElementString("Name", agent.strName);
				save.WriteElementString("GamesPlayed", agent.iGames.ToString());
				save.WriteElementString("Fitness", agent.dFitness.ToString());
				save.Indentation--;
				save.WriteEndElement();
				// write program
				save.WriteStartElement("Program");
				save.Indentation++;
				if (agent.strAssembly.Length > 0)
					save.WriteElementString("Assembly", agent.strAssembly);
				for (int i = 0; i < ExchangeAgent_Programmed.iProgramSize; i++) {
					int opcode = (int) agent.cProgram[i];
					save.WriteElementString("Opcode", opcode.ToString());
				}
				save.Indentation--;
				save.WriteEndElement();
				save.Indentation--;
				save.WriteEndElement();
			}
			// close XML file
			save.Indentation--;
			save.WriteEndElement();
			save.WriteEndDocument();
			save.Close();
		}

		public static ArrayList LoadSet(string file) {
			// This function loads an ExchangeAgent set from disk. This file explicitly relies on the ExchangeAgentSet.xsd definition.
			ArrayList agents = new ArrayList();
			try {
				XmlTextReader xtr = new XmlTextReader(file);
				XmlValidatingReader xvr = new XmlValidatingReader(xtr);
				xvr.ValidationType = ValidationType.Schema;
				xvr.ValidationEventHandler += new ValidationEventHandler(ValidationErrors);
				XmlSchemaCollection xsc = new XmlSchemaCollection();
				xsc.Add("", System.Reflection.Assembly.GetExecutingAssembly().Location.Substring(0, System.Reflection.Assembly.GetExecutingAssembly().Location.Length - 16) + @"4 Resources\ExchangeAgentSet.xsd");
				xvr.Schemas.Add(xsc);
				XmlDocument xd = new XmlDocument();
				XmlLoadErrors = "";
				xd.Load(xvr);
				if (XmlLoadErrors.Length > 0)
					MessageBox.Show("Errors loading network document. This does not appear to be a valid Exchange agent.", "Errors Loading Exchange Agent", MessageBoxButtons.OK, MessageBoxIcon.Error);
				else {
					foreach (XmlNode x1 in xd.ChildNodes) {
						if (x1.Name == "ExchangeAgentSet") {
							foreach (XmlNode x2 in x1.ChildNodes) {
								if (x2.Name == "ExchangeAgent") {
									ExchangeAgent_Programmed agent = new ExchangeAgent_Programmed(new Random(0));
									foreach (XmlNode x3 in x2.ChildNodes) {
										if (x3.Name == "Parameters") {
											foreach (XmlNode x4 in x3.ChildNodes) {
												if (x4.Name == "Name")
													agent.strName = x4.InnerText;
												else if (x4.Name == "GamesPlayed")
													agent.iGames = Int32.Parse(x4.InnerText);
												else if (x4.Name == "Fitness")
													agent.dFitness = double.Parse(x4.InnerText);
											}
										}
										else if (x3.Name == "Program") {
											int count = 0;
											foreach (XmlNode x4 in x3.ChildNodes) {
												if (x4.Name == "Assembly")
													agent.strAssembly = x4.InnerText;
												else if (x4.Name == "Opcode")
													agent.cProgram[count++] = (char) Int32.Parse(x4.InnerText);
											}
											agents.Add(agent);
										}
									}
								}
							}
						}
					}
					xtr.Close();
					return agents;
				}
			} catch (Exception ex) {
				MessageBox.Show(ex.Message);
				return null;
			}
			return null;
		}

		public static void ValidationErrors(object sender, ValidationEventArgs e) {
			XmlLoadErrors += e.Message + "\r\n";
		}

		public ExchangeAgent_Programmed Clone() {
			ExchangeAgent_Programmed newAgent = new ExchangeAgent_Programmed();
			newAgent.dFitness = dFitness;
			newAgent.dRank = dRank;
			newAgent.exchange = exchange;
			newAgent.iBuyGood = iBuyGood;
			newAgent.iGames = iGames;
			newAgent.iProduction = iProduction;
			newAgent.iSellGood = iSellGood;
			newAgent.lValue = lValue;
			newAgent.player = player;
			newAgent.strName = strName;
			newAgent.strAssembly = strAssembly;
			for (int i = 0; i < ExchangeAgent_Programmed.iProgramSize; i++)
				newAgent.cProgram[i] = cProgram[i];
			return newAgent;
		}

		#endregion

	}

	public class ExchangeVirtualMachine {
		// This class represents the virtual machine in which the program of an ExchangeAgent executes.

		#region Variables and initialization code

		int iProgramCounter;
		int iTimer;
		short[] sRegister;
		bool bProgramDone;

		public ExchangeVirtualMachine() {
			sRegister = new short[4];
		}

		#endregion

		#region Core functions

		public void Execute(ExchangeAgent_Programmed agent) {
			// This function executes the ExchangeAgent program.
			// initialize machine
			iProgramCounter = 0;
			sRegister[0] = sRegister[1] = sRegister[2] = sRegister[3] = 0;
			iTimer = 0;
			bProgramDone = false;
			// execute instructions
			while (bProgramDone == false) {
				int cInstruction = agent.cProgram[iProgramCounter];
				iProgramCounter++;

				// parse instruction
				switch (cInstruction) {

						// NOOP instruction
					case 0:  // NOOP
						break;

						// Set Register instructions
					case 1:
					case 2:
					case 3:
					case 4:  // set register to zero
						sRegister[cInstruction - 1] = 0;
						break;
					case 5:
					case 6:
					case 7:
					case 8:  // set register to fixed constant
						if (iProgramCounter < ExchangeAgent_Programmed.iProgramSize) {
							sRegister[cInstruction - 5] = (agent.cProgram[iProgramCounter] > 128 ? (short) ((int) agent.cProgram[iProgramCounter] - 256) : (short) agent.cProgram[iProgramCounter]);
							iProgramCounter++;
						}
						break;
					case 9:
					case 10:
					case 11:
					case 12:
					case 13:
					case 14:
					case 15:
					case 16:
					case 17:
					case 18:
					case 19:
					case 20: {  // set register to register
						int iDestRegister = (cInstruction - 9) / 3;
						int iSrcRegister = (cInstruction - 9) % 3;
						if (iSrcRegister >= iDestRegister)
							iSrcRegister++;
						sRegister[iDestRegister] = sRegister[iSrcRegister];
						break;
					}
					case 21:
					case 22:
					case 23:
					case 24:
					case 25:
					case 26:
					case 27:
					case 28:
					case 29:
					case 30:
					case 31:
					case 32: {  // swap registers
						int iDestRegister = (cInstruction - 21) / 3;
						int iSrcRegister = (cInstruction - 21) % 3;
						if (iSrcRegister >= iDestRegister)
							iSrcRegister++;
						short swap = sRegister[iDestRegister];
						sRegister[iDestRegister] = sRegister[iSrcRegister];
						sRegister[iSrcRegister] = swap;
						break;
					}

						// Load Variables Into Accumulator
					case 33:
					case 34:
					case 35:
					case 36:
					case 37:
					case 38:
					case 39:
					case 40:
					case 41:  // load resource price into accumulator (Register 0)
						sRegister[0] = (short) ((Good) agent.exchange.goods[cInstruction - 33]).iPrice;
						break;
					case 42:
					case 43:
					case 44:
					case 45:
					case 46:
					case 47:
					case 48:
					case 49:
					case 50:  // load resource price trend into accumulator (Register 0)
						sRegister[0] = (short) ((Good) agent.exchange.goods[cInstruction - 42]).iTrend;
						break;
					case 51:
					case 52:
					case 53:
					case 54:
					case 55:
					case 56:
					case 57:
					case 58:
					case 59:  // load resource quantity into accumulator (Register 0)
						sRegister[0] = (short) ((Good) agent.exchange.goods[cInstruction - 51]).iQuantity;
						break;
					case 60:
					case 61:
					case 62:
					case 63:
					case 64:
					case 65:
					case 66:
					case 67:
					case 68:  // load resource holdings into accumulator (Register 0)
						sRegister[0] = (short) agent.player.iHoldings[cInstruction - 60];
						break;
					case 69:  // load player credits into accumulator (Register 0)
						sRegister[0] = (short) agent.player.iCredits;
						break;
					case 70:  // load event into accumulator (Register 0)
						sRegister[0] = (short) agent.exchange.eEvent;
						break;
					case 71:  // load phase timer into accumulator (Register 0)
						sRegister[0] = (short) agent.exchange.iPhase;
						break;
					case 72:  // load player count into accumulator (Register 0)
						sRegister[0] = (short) agent.exchange.agents.Count;
						break;
					case 73:  // load random number into accumulator (Register 0)
						if (sRegister[0] > 0)
							sRegister[0] = (short) agent.exchange.random.Next(sRegister[0] - 1);
						break;

						// Register Math
					case 74:
					case 75:
					case 76:
					case 77:  // add constant to register
						if (iProgramCounter < ExchangeAgent_Programmed.iProgramSize) {
							sRegister[cInstruction - 74] += (short) agent.cProgram[iProgramCounter];
							iProgramCounter++;
						}
						break;
					case 78:
					case 79:
					case 80:
					case 81:
					case 82:
					case 83:
					case 84:
					case 85:
					case 86:
					case 87:
					case 88:
					case 89: {  // add register to register
						int iDestRegister = (cInstruction - 78) / 3;
						int iSrcRegister = (cInstruction - 78) % 3;
						if (iSrcRegister >= iDestRegister)
							iSrcRegister++;
						sRegister[iDestRegister] += sRegister[iSrcRegister];
						break;
					}
					case 90:
					case 91:
					case 92:
					case 93: {  // subtract constant from register
						if (iProgramCounter < ExchangeAgent_Programmed.iProgramSize) {
							sRegister[cInstruction - 90] -= (short) agent.cProgram[iProgramCounter];
							iProgramCounter++;
						}
						break;
					}
					case 94:
					case 95:
					case 96:
					case 97:
					case 98:
					case 99:
					case 100:
					case 101:
					case 102:
					case 103:
					case 104:
					case 105: {  // subtract register from register
						int iDestRegister = (cInstruction - 94) / 3;
						int iSrcRegister = (cInstruction - 94) % 3;
						if (iSrcRegister >= iDestRegister)
							iSrcRegister++;
						sRegister[iDestRegister] -= sRegister[iSrcRegister];
						break;
					}
					case 106:
					case 107:
					case 108:
					case 109:  // multiply register by contstant
						if (iProgramCounter < ExchangeAgent_Programmed.iProgramSize) {
							sRegister[cInstruction - 106] *= (short) agent.cProgram[iProgramCounter];
							iProgramCounter++;
						}
						break;
					case 110:
					case 111:
					case 112:
					case 113:
					case 114:
					case 115:
					case 116:
					case 117:
					case 118:
					case 119:
					case 120:
					case 121: {  // multiply register by register
						int iDestRegister = (cInstruction - 110) / 3;
						int iSrcRegister = (cInstruction - 110) % 3;
						if (iSrcRegister >= iDestRegister)
							iSrcRegister++;
						sRegister[iDestRegister] *= sRegister[iSrcRegister];
						break;
					}
					case 122:
					case 123:
					case 124:
					case 125:  // divide register by constant
						if ((iProgramCounter < ExchangeAgent_Programmed.iProgramSize) && (agent.cProgram[iProgramCounter] != 0)) {
							sRegister[cInstruction - 122] /= (short) agent.cProgram[iProgramCounter];
							iProgramCounter++;
						}
						break;
					case 126:
					case 127:
					case 128:
					case 129:
					case 130:
					case 131:
					case 132:
					case 133:
					case 134:
					case 135:
					case 136:
					case 137: {  // divide register by register
						int iDestRegister = (cInstruction - 126) / 3;
						int iSrcRegister = (cInstruction - 126) % 3;
						if (iSrcRegister >= iDestRegister)
							iSrcRegister++;
						if (sRegister[iSrcRegister] != 0)
							sRegister[iDestRegister] /= sRegister[iSrcRegister];
						break;
					}

						// Jump Instructions (Non-Conditional)
					case 138:  // unconditional jump instructions
						if (iProgramCounter < ExchangeAgent_Programmed.iProgramSize - 1) {
							int iTempProgramCounter = (agent.cProgram[iProgramCounter] > 128 ? iProgramCounter - 257 + agent.cProgram[iProgramCounter] : iProgramCounter + agent.cProgram[iProgramCounter] + 1);
							if ((iTempProgramCounter >= 0) && (iTempProgramCounter < ExchangeAgent_Programmed.iProgramSize))
								iProgramCounter = iTempProgramCounter;
							else  // if error, skip over constant
								iProgramCounter++;
						}
						break;
					case 139:
					case 140:
					case 141:
					case 142:  // jump (constant) instructions if register = 0
						if ((iProgramCounter < ExchangeAgent_Programmed.iProgramSize - 1) && (sRegister[cInstruction - 139] == 0)) {
							int iTempProgramCounter = (agent.cProgram[iProgramCounter] > 128 ? iProgramCounter - 257 + agent.cProgram[iProgramCounter] : iProgramCounter + agent.cProgram[iProgramCounter] + 1);
							if ((iTempProgramCounter >= 0) && (iTempProgramCounter < ExchangeAgent_Programmed.iProgramSize))
								iProgramCounter = iTempProgramCounter;
							else  // if error, skip over constant
								iProgramCounter++;
						}
						else
							iProgramCounter++;  // jump over trailing constant
						break;
					case 143:
					case 144:
					case 145:
					case 146:  // jump forward (constant) instructions if register != 0
						if ((iProgramCounter < ExchangeAgent_Programmed.iProgramSize - 1) && (sRegister[cInstruction - 143] != 0)) {  // need to jump forward at least two instructions
							int iTempProgramCounter = (agent.cProgram[iProgramCounter] > 128 ? iProgramCounter - 257 + agent.cProgram[iProgramCounter] : iProgramCounter + agent.cProgram[iProgramCounter] + 1);
							if ((iTempProgramCounter >= 0) && (iTempProgramCounter < ExchangeAgent_Programmed.iProgramSize))
								iProgramCounter = iTempProgramCounter;
							else  // if error, skip over constant
								iProgramCounter++;
						}
						else
							iProgramCounter++;  // jump over trailing constant
						break;

						// Conditional Instructions (either execute immediately following instruction or not)
					case 147:
					case 148:
					case 149:
					case 150:  // compare register to constant (==)
						if (iProgramCounter < ExchangeAgent_Programmed.iProgramSize - 1) {
							if (sRegister[cInstruction - 147] == agent.cProgram[iProgramCounter])
								iProgramCounter++;  // execute next instruction
							else  // skip next instruction (if it's a variable-length instruction, skip three; otherwise skip two)
								iProgramCounter += (((iProgramCounter < ExchangeAgent_Programmed.iProgramSize - 1) && (ExtendedOpcode(agent.cProgram[iProgramCounter + 1]))) ? 3 : 2);
						}
						else
							iProgramCounter++;
						break;
					case 151:
					case 152:
					case 153:
					case 154:
					case 155:
					case 156:
					case 157:
					case 158:
					case 159:
					case 160:
					case 161:
					case 162:  // compare register to register (==)
						if (iProgramCounter < ExchangeAgent_Programmed.iProgramSize - 1) {
							int iDestRegister = (cInstruction - 151) / 3;
							int iSrcRegister = (cInstruction - 151) % 3;
							if (iSrcRegister >= iDestRegister)
								iSrcRegister++;
							if (sRegister[iSrcRegister] == sRegister[iDestRegister])
								iProgramCounter = iProgramCounter;  // execute next instruction
							else  // skip next instruction (if it's a variable-length instruction, skip two; otherwise skip one)
								iProgramCounter += (ExtendedOpcode(agent.cProgram[iProgramCounter + 1]) ? 2 : 1);
						}
						break;
					case 163:
					case 164:
					case 165:
					case 166:  // compare register to constant (!=)
						if (iProgramCounter < ExchangeAgent_Programmed.iProgramSize - 1) {
							if (sRegister[cInstruction - 163] != agent.cProgram[iProgramCounter])
								iProgramCounter++;  // execute next instruction
							else  // skip next instruction (if it's a variable-length instruction, skip three; otherwise skip two)
								iProgramCounter += (((iProgramCounter < ExchangeAgent_Programmed.iProgramSize - 1) && (ExtendedOpcode(agent.cProgram[iProgramCounter + 1]))) ? 3 : 2);
						}
						else
							iProgramCounter++;
						break;
					case 167:
					case 168:
					case 169:
					case 170:
					case 171:
					case 172:
					case 173:
					case 174:
					case 175:
					case 176:
					case 177:
					case 178:  // compare register to register (!=)
						if (iProgramCounter < ExchangeAgent_Programmed.iProgramSize - 1) {
							int iDestRegister = (cInstruction - 167) / 3;
							int iSrcRegister = (cInstruction - 167) % 3;
							if (iSrcRegister >= iDestRegister)
								iSrcRegister++;
							if (sRegister[iSrcRegister] != sRegister[iDestRegister])
								iProgramCounter = iProgramCounter;  // execute next instruction
							else  // skip next instruction (if it's a variable-length instruction, skip two; otherwise skip one)
								iProgramCounter += (ExtendedOpcode(agent.cProgram[iProgramCounter + 1]) ? 2 : 1);
						}
						break;
					case 179:
					case 180:
					case 181:
					case 182:  // compare register to constant (>)
						if (iProgramCounter < ExchangeAgent_Programmed.iProgramSize - 1) {
							if (sRegister[cInstruction - 179] > agent.cProgram[iProgramCounter])
								iProgramCounter++;  // execute next instruction
							else  // skip next instruction (if it's a variable-length instruction, skip three; otherwise skip two)
								iProgramCounter += (((iProgramCounter < ExchangeAgent_Programmed.iProgramSize - 1) && (ExtendedOpcode(agent.cProgram[iProgramCounter + 1]))) ? 3 : 2);
						}
						else
							iProgramCounter++;
						break;
					case 183:
					case 184:
					case 185:
					case 186:
					case 187:
					case 188:
					case 189:
					case 190:
					case 191:
					case 192:
					case 193:
					case 194:  // compare register to register (>)
						if (iProgramCounter < ExchangeAgent_Programmed.iProgramSize - 1) {
							int iDestRegister = (cInstruction - 183) / 3;
							int iSrcRegister = (cInstruction - 183) % 3;
							if (iSrcRegister >= iDestRegister)
								iSrcRegister++;
							if (sRegister[iSrcRegister] > sRegister[iDestRegister])
								iProgramCounter = iProgramCounter;  // execute next instruction
							else  // skip next instruction (if it's a variable-length instruction, skip two; otherwise skip one)
								iProgramCounter += (ExtendedOpcode(agent.cProgram[iProgramCounter + 1]) ? 2 : 1);
						}
						break;
					case 195:
					case 196:
					case 197:
					case 198:  // compare register to constant (>=)
						if (iProgramCounter < ExchangeAgent_Programmed.iProgramSize - 2) {
							if (sRegister[cInstruction - 195] >= agent.cProgram[iProgramCounter])
								iProgramCounter++;  // execute next instruction
							else  // skip next instruction (if it's a variable-length instruction, skip three; otherwise skip two)
								iProgramCounter += (ExtendedOpcode(agent.cProgram[iProgramCounter + 1]) ? 3 : 2);
						}
						else
							iProgramCounter++;
						break;
					case 199:
					case 200:
					case 201:
					case 202:
					case 203:
					case 204:
					case 205:
					case 206:
					case 207:
					case 208:
					case 209:
					case 210:  // compare register to register (>=)
						if (iProgramCounter < ExchangeAgent_Programmed.iProgramSize - 1) {
							int iDestRegister = (cInstruction - 199) / 3;
							int iSrcRegister = (cInstruction - 199) % 3;
							if (iSrcRegister >= iDestRegister)
								iSrcRegister++;
							if (sRegister[iDestRegister] >= sRegister[iSrcRegister])
								iProgramCounter = iProgramCounter;  // execute next instruction
							else  // skip next instruction (if it's a variable-length instruction, skip two; otherwise skip one)
								iProgramCounter += (ExtendedOpcode(agent.cProgram[iProgramCounter + 1]) ? 2 : 1);
						}
						break;
					case 211:
					case 212:
					case 213:
					case 214:  // compare register to constant (<)
						if (iProgramCounter < ExchangeAgent_Programmed.iProgramSize - 1) {
							if (sRegister[cInstruction - 211] < agent.cProgram[iProgramCounter])
								iProgramCounter++;  // execute next instruction
							else  // skip next instruction (if it's a variable-length instruction, skip three; otherwise skip two)
								iProgramCounter += (((iProgramCounter < ExchangeAgent_Programmed.iProgramSize - 1) && (ExtendedOpcode(agent.cProgram[iProgramCounter + 1]))) ? 3 : 2);
						}
						else
							iProgramCounter++;
						break;
					case 215:
					case 216:
					case 217:
					case 218:
					case 219:
					case 220:
					case 221:
					case 222:
					case 223:
					case 224:
					case 225:
					case 226:  // compare register to register (<)
						if (iProgramCounter < ExchangeAgent_Programmed.iProgramSize - 1) {
							int iDestRegister = (cInstruction - 215) / 3;
							int iSrcRegister = (cInstruction - 215) % 3;
							if (iSrcRegister >= iDestRegister)
								iSrcRegister++;
							if (sRegister[iDestRegister] < sRegister[iSrcRegister])
								iProgramCounter = iProgramCounter;  // execute next instruction
							else  // skip next instruction (if it's a variable-length instruction, skip two; otherwise skip one)
								iProgramCounter += (ExtendedOpcode(agent.cProgram[iProgramCounter + 1]) ? 2 : 1);
						}
						break;
					case 227:
					case 228:
					case 229:
					case 230:  // compare register to constant (<=)
						if (iProgramCounter < ExchangeAgent_Programmed.iProgramSize - 1) {
							if (sRegister[cInstruction - 227] <= agent.cProgram[iProgramCounter])
								iProgramCounter++;  // execute next instruction
							else  // skip next instruction (if it's a variable-length instruction, skip three; otherwise skip two)
								iProgramCounter += (((iProgramCounter < ExchangeAgent_Programmed.iProgramSize - 1) && (ExtendedOpcode(agent.cProgram[iProgramCounter + 1]))) ? 3 : 2);
						}
						else
							iProgramCounter++;
						break;
					case 231:
					case 232:
					case 233:
					case 234:
					case 235:
					case 236:
					case 237:
					case 238:
					case 239:
					case 240:
					case 241:
					case 242:  // compare register to register (<=)
						if (iProgramCounter < ExchangeAgent_Programmed.iProgramSize - 1) {
							int iDestRegister = (cInstruction - 231) / 3;
							int iSrcRegister = (cInstruction - 231) % 3;
							if (iSrcRegister >= iDestRegister)
								iSrcRegister++;
							if (sRegister[iDestRegister] <= sRegister[iSrcRegister])
								iProgramCounter = iProgramCounter;  // execute next instruction
							else  // skip next instruction (if it's a variable-length instruction, skip two; otherwise skip one)
								iProgramCounter += (ExtendedOpcode(agent.cProgram[iProgramCounter + 1]) ? 2 : 1);
						}
						break;

						// Store Registers as Actions
					case 243:
					case 244:
					case 245:
					case 246:  // store register as Buy order
						if ((sRegister[cInstruction - 243] >= -1) && (sRegister[cInstruction - 243] <= 8))
							agent.iBuyGood = sRegister[cInstruction - 243];
						break;
					case 247:
					case 248:
					case 249:
					case 250:  // store register as Buy order
						if ((sRegister[cInstruction - 247] >= -1) && (sRegister[cInstruction - 247] <= 8))
							agent.iSellGood = sRegister[cInstruction - 247];
						break;
					case 251:
					case 252:
					case 253:
					case 254:  // store register as Production order
						if ((sRegister[cInstruction - 251] >= 0) && (sRegister[cInstruction - 251] <= 8))
							agent.iProduction = sRegister[cInstruction - 251];
						break;

						// Finished Instructions
					case 255:  // end
						bProgramDone = true;
						break;

					default:  // this code should be unreachable
						break;
				}
				// check termination conditions
				if (((++iTimer) >= ExchangeAgent_Programmed.iProgramTimeout) || (iProgramCounter < 0) || (iProgramCounter >= ExchangeAgent_Programmed.iProgramSize))
					bProgramDone = true;
			}
		}

		bool ExtendedOpcode(char instruction) {   // return true if this is an extended opcode, or false otherwise
			if (((instruction >= 5) && (instruction <= 8)) ||
				((instruction >= 74) && (instruction <= 77)) ||
				((instruction >= 90) && (instruction <= 93)) ||
				((instruction >= 106) && (instruction <= 109)) ||
				((instruction >= 122) && (instruction <= 125)) ||
				((instruction >= 138) && (instruction <= 150)) ||
				((instruction >= 163) && (instruction <= 166)) ||
				((instruction >= 179) && (instruction <= 182)) ||
				((instruction >= 195) && (instruction <= 198)) ||
				((instruction >= 211) && (instruction <= 214)) ||
				((instruction >= 227) && (instruction <= 230)))
				return true;
			else
				return false;
		}

		#endregion

	}
}
