// The Exchange: Genetically Evolving Market Agents 
// Written by David J. Stein, Esq., 11/22/2004. See "License.html" for terms and conditions of use.
//
// This module contains all of the code for the Exchange simulation, including the Good classes and the Player class.

#region Using declarations

using System;
using System.Collections;

#endregion

namespace The_Exchange_Namespace {

	public class ExchangeSimulation {
		// This class contains the code for running the Exchange simulation.

		#region Variables and initialization code

		public ArrayList goods;							// an array of Good objects, each representing an exchanged good
		public ArrayList agents;						// an array of ExchangeAgent objects, each representing (wrapping) a Player object
		public Random random;						// a randomizer
		public int iPhase, iTimer;						// the timer objects - iTimer ticks up per second from 0 to 60; iPhase records minutes from 1 to 10
		public bool bRunning, bDone;		// booleans describing whehter or not the simulation is running, and whether or not it is complete
		public string strNews;							// string containing the current line running across the news ticker
		public string strEndEvent;					// string to be appended to strNews when the current event ends
		public int iEventDuration;					// the duration of the current event
		public EventsEnum eEvent;				// the enum of the current event
		public enum EventsEnum { None, Pirates, Fire, Convoy, Ruins, Lottery, Tax, Subsidy, MerchantConvoy, GoodEconomy, Cooperation,
			SolarFlare, DustStorm, SludgePool, OreVein, SpiceHarvest, RampantWharks, WharkBreeding, SaveTheWharks, SyrineCave,
			Planetquakes, MardiGras, DistilleryClosed	};			// an enum describing all of the current events

		public ExchangeSimulation(Random r) {
			random = r;
			// create goods
			goods = new ArrayList();
			goods.Add(new Good_Power(this, 20, 25));
			goods.Add(new Good_Ore(this, 20, 25));
			goods.Add(new Good_Sludge(this, 20, 25));
			goods.Add(new Good_Galactite(this, 50, 0));
			goods.Add(new Good_Hydradine(this, 50, 0));
			goods.Add(new Good_Spice(this, 50, 0));
			goods.Add(new Good_Shells(this, 50, 0));
			goods.Add(new Good_Crystals(this, 50, 0));
			goods.Add(new Good_Rum(this, 50, 0));
			foreach (Good good in goods)  // set initial prices
				good.SetPrice(false);
			agents = new ArrayList();
			iPhase = 1;
			iTimer = 0;
			eEvent = EventsEnum.None;
			iEventDuration = 0;
			bRunning = false;
			bDone = false;
		}

		#endregion

		#region Core functions

		public void Step() {
			// This function runs one turn in the simulation. This function is called once per second while the game is running.
			ExecuteRandomEvent();  // choose a random event
			BuildGoods();							// have the store build goods
			AdjustGoodPrices();			// adjust the prices of all goods
			ProcessPlayers();					// allow players to build goods and process commands (basically just for the programmed artificial intelligence agents)
			SetValues();								// set the current values and rankings of all players
			UpdateTimer();						// update the game timer, and end the game if time is up
		}

		public void ExecuteRandomEvent() {
			// This function handles random event.
			if (eEvent != EventsEnum.None) {
				if ((--iEventDuration) <= 0) {
					eEvent = EventsEnum.None;
					iEventDuration = 0;
					strNews += strEndEvent;
					strEndEvent = "";
				}
			}
			else if (random.Next(60) == 0) {
				int iEvent = random.Next(22) + 1;
				switch (iEvent) {
					case 1:
						eEvent = EventsEnum.Pirates;
						int iGood = random.Next(6) + 3;
						strNews += " * Pirates Raid Colony - Seize All " + ((Good) goods[iGood]).strName + " * ";
						((Good) goods[iGood]).iQuantity = 0;
						foreach (ExchangeAgent agent in agents)
							agent.player.iHoldings[iGood] = 0;
						iEventDuration = 30;
						break;
					case 2:
						eEvent = EventsEnum.Fire;
						strNews += " * Colony Fire - All Store Quantities Lost * ";
						foreach (Good g in goods)
							g.iQuantity = 0;
						iEventDuration = 30;
						break;
					case 3:
						eEvent = EventsEnum.Convoy;
						strNews += " * Imperial Convoy Resupplies Colony * ";
						((Good) goods[(int) Good.GoodEnum.Power]).iQuantity += 30;
						((Good) goods[1]).iQuantity += 30;
						((Good) goods[2]).iQuantity += 30;
						iEventDuration = 30;
						break;
					case 4:
						eEvent = EventsEnum.Ruins;
						strNews += " * Colony Discovers (Plunders) Ancient Ruins - Precious Goods Increased * ";
						((Good) goods[5]).iQuantity += 30;
						((Good) goods[6]).iQuantity += 30;
						((Good) goods[7]).iQuantity += 30;
						((Good) goods[8]).iQuantity += 30;
						iEventDuration = 30;
						break;
					case 5:
						eEvent = EventsEnum.Lottery;
						int player = random.Next(agents.Count);
						int winnings = random.Next(6) * 50 + 200;
						int[] lottery = new int[5];
						for (int i = 0; i < 5; i++)
							lottery[i] = random.Next(30) + 1;
						strNews += " * Colony Lottery Results: Winning Ticket = " + lottery[0] + " " + lottery[1] + " " + lottery[2] + " " + lottery[3] + " " + lottery[4] + ": " + ((ExchangeAgent) agents[player]).player.strName + " Wins " + winnings + " Credits * ";
						((ExchangeAgent) agents[player]).player.iCredits += winnings;
						iEventDuration = 30;
						break;
					case 6:
						eEvent = EventsEnum.Tax;
						int tax = iPhase * 15 + 20;
						foreach (ExchangeAgent agent in agents)
							agent.player.iCredits = Math.Max(0, agent.player.iCredits - tax);
						strNews += " * Galactic Chancellor Levies Tax of " + tax + " Credits * ";
						iEventDuration = 30;
						break;
					case 7:
						eEvent = EventsEnum.Subsidy;
						int credits = random.Next(10) * (5 + iPhase) + 100;
						foreach (ExchangeAgent agent in agents)
							agent.player.iCredits += credits;
						strNews += " * Galactic Subsidizes Colonization - All Players Receive " + credits + " Credits * ";
						iEventDuration = 30;
						break;
					case 8:
						eEvent = EventsEnum.MerchantConvoy;
						strNews += " * Merchant Convoy Visits System To Purchase Exotic Supplies - All Prices Doubled * ";
						strEndEvent = " * Merchant Convoy Departs, Prices Return To Normal * ";
						iEventDuration = 60;
						break;
					case 9:
						eEvent = EventsEnum.GoodEconomy;
						strNews += " * Healthy Economy Raises Value of Galactic Credit - All Prices Decreased * ";
						strEndEvent = " * Galactic Economy Collapses - Prices Deflate To Pre-Boom Levels * ";
						iEventDuration = 60;
						break;
					case 10:
						eEvent = EventsEnum.Cooperation;
						strNews += " * Colony Community Feeling Cooperative - All Productivity Doubled * ";
						strEndEvent = " * Colony Dispute Breaks Out - Productivity Diminished * ";
						iEventDuration = 60;
						break;
					case 11:
						eEvent = EventsEnum.SolarFlare;
						strNews += " * Solar Flare - Energy Production Tripled * ";
						strEndEvent = " * Solar Flare Dissipates * ";
						iEventDuration = 60;
						break;
					case 12:
						eEvent = EventsEnum.DustStorm;
						strNews += " * Dust Storm - Energy Production Halved * ";
						strEndEvent = " * Dust Storm Blows Over - Energy Production Returns To Normal * ";
						iEventDuration = 60;
						break;
					case 13:
						eEvent = EventsEnum.SludgePool;
						strNews += " *  New Sludge Pool Discovered - Production Doubled * ";
						strEndEvent = " * Sludge Pool Tapped Out * ";
						iEventDuration = 60;
						break;
					case 14:
						eEvent = EventsEnum.OreVein;
						strNews += " *  Ore Vein Discovered - Production Doubled * ";
						strEndEvent = " * Excess Mining Exhausts New Ore Vein * ";
						iEventDuration = 60;
						break;
					case 15:
						eEvent = EventsEnum.SpiceHarvest;
						strNews += " * Spice Harvest Season Begins - Production Doubled * ";
						strEndEvent = " * Spice Harvest Ends * ";
						iEventDuration = 60;
						break;
					case 16:
						eEvent = EventsEnum.RampantWharks;
						strNews += " * Rampant Wharks Decimates Crops - Spice Production Halved * ";
						strEndEvent = " * Whark Population Thinned - Spice Production Returns to Normal Levels * ";
						iEventDuration = 60;
						break;
					case 17:
						eEvent = EventsEnum.WharkBreeding;
						strNews += " * Whark Breeding Out of Control - Colony Raises Whark Shell Price to Encourage Hunting * ";
						strEndEvent = " * Whark Population Under Control - Shell Incentive Program Ends * ";
						iEventDuration = 60;
						break;
					case 18:
						eEvent = EventsEnum.SaveTheWharks;
						strNews += " * Animal Rights Group Initiates 'Save-the-Wharks' Campaign - Whark Shell Demand Falls * ";
						strEndEvent = " * Animal Rights Group Leaves Quadrant - Shell Demand Returns To Normal * ";
						iEventDuration = 60;
						break;
					case 19:
						eEvent = EventsEnum.SyrineCave;
						strNews += " * Spelunkers Find Massive Cave Complex - Syrine Crystal Production Up * ";
						strEndEvent = " * Cave-In Blocks Syrine Crystal Cavern - Production Returns To Normal * ";
						iEventDuration = 60;
						break;
					case 20:
						eEvent = EventsEnum.Planetquakes;
						strNews += " * Planetquakes Hamper Spelunking Efforts - Syrine Crystal Production Diminished * ";
						strEndEvent += " * Planetquakes End - Syrine Crystal Production Resumes * ";
						iEventDuration = 60;
						break;
					case 21:
						eEvent = EventsEnum.MardiGras;
						strNews += " * Quadrant Celebrates Mardi Gras - Denebian Rum Production Doubles * ";
						strEndEvent += " * Mardi Gras Ends - Denebian Rum Demand Nominal * ";
						iEventDuration = 60;
						break;
					case 22:
						eEvent = EventsEnum.DistilleryClosed;
						strNews += " * Distillery Closes For Renovation - Denebian Rum Production Halts * ";
						strEndEvent += " * Distillery Reopens - Denebian Rum Production Resumes * ";
						iEventDuration = 60;
						break;
					default:  // do nothing
						break;
				}
			}
		}

		public void BuildGoods() {
			// This function carries out production for the market.
			int iProduction = (eEvent == EventsEnum.Cooperation ? 2 : 1);
			if (iTimer % 2 == 1) {  // produce power units every other turn
				if (eEvent == EventsEnum.SolarFlare)
					iProduction = 3;
				else if ((eEvent == EventsEnum.DustStorm) && (random.Next(2) == 0))
					iProduction = 0;
				int units = random.Next(6);
				if (units > 0)
					((Good) goods[(int) Good.GoodEnum.Power]).iQuantity += iProduction + (units == 5 ? iProduction : 0);
			}
			else if (((Good) goods[(int) Good.GoodEnum.Power]).iQuantity >= iProduction) {
				int good = (iTimer / 2) % 8;
				switch (good) {
					case 0:  // produce ore
						if (eEvent == EventsEnum.OreVein)
							iProduction = 2;
						((Good) goods[1]).iQuantity += iProduction;
						((Good) goods[(int) Good.GoodEnum.Power]).iQuantity -= iProduction;
						break;
					case 1:  // produce sludge
						if (eEvent == EventsEnum.SludgePool)
							iProduction = 2;
						((Good) goods[2]).iQuantity += iProduction;
						((Good) goods[(int) Good.GoodEnum.Power]).iQuantity -= iProduction;
						break;
					case 2:  // produce galactite if we have a unit of ore
						if (((Good) goods[1]).iQuantity >= iProduction) {
							((Good) goods[3]).iQuantity += iProduction;
							((Good) goods[1]).iQuantity -= iProduction;
							((Good) goods[(int) Good.GoodEnum.Power]).iQuantity -= iProduction;
						}
						break;
					case 3:  // produce hydradine if we have a unit of sludge
						if (((Good) goods[2]).iQuantity >= iProduction) {
							((Good) goods[4]).iQuantity += iProduction;
							((Good) goods[2]).iQuantity -= iProduction;
							((Good) goods[(int) Good.GoodEnum.Power]).iQuantity -= iProduction;
						}
						break;
					case 4:  // produce spice
						if (eEvent == EventsEnum.SpiceHarvest)
							iProduction = 2;
						else if ((eEvent == EventsEnum.RampantWharks) && (random.Next(2) == 0))
							iProduction = 0;
						if (random.Next(3) > 0) {
							((Good) goods[5]).iQuantity += iProduction;
							((Good) goods[(int) Good.GoodEnum.Power]).iQuantity -= iProduction;
						}
						break;
					case 5:  // produce shells
						if (eEvent == EventsEnum.WharkBreeding)
							iProduction = 2;
						else if ((eEvent == EventsEnum.SaveTheWharks) && (random.Next(2) == 0))
							iProduction = 0;
						if (random.Next(3) > 0) {
							((Good) goods[6]).iQuantity += iProduction;
							((Good) goods[(int) Good.GoodEnum.Power]).iQuantity -= iProduction;
						}
						break;
					case 6:  // produce crystals
						if (eEvent == EventsEnum.SyrineCave)
							iProduction = 2;
						else if ((eEvent == EventsEnum.Planetquakes) && (random.Next(2) == 0))
							iProduction = 0;
						if (random.Next(3) > 0) {
							((Good) goods[7]).iQuantity += iProduction;
							((Good) goods[(int) Good.GoodEnum.Power]).iQuantity -= iProduction;
						}
						break;
					case 7:  // produce rum
						if (eEvent == EventsEnum.MardiGras)
							iProduction = 2;
						else if (eEvent == EventsEnum.DistilleryClosed)
							iProduction = 0;
						if (random.Next(3) > 0) {
							((Good) goods[8]).iQuantity += iProduction;
							((Good) goods[(int) Good.GoodEnum.Power]).iQuantity -= iProduction;
						}
						break;
				}
			}
		}

		public void AdjustGoodPrices() {
			// This function sets the price of each good, according to its specific pricing function.
			foreach (Good g in goods)
				g.SetPrice(true);
			// adjust pricehistory[0] to avoid overlap, in order to display all lines on chart
			for (int i =1; i < goods.Count; i++) {
				for (int j = 0; j < i; j++) {
					if (((Good) goods[i]).iPriceHistory[0] == ((Good) goods[j]).iPriceHistory[0]) {
						((Good) goods[i]).iPriceHistory[0]++;
						j = -1;  // reset counter to start over
					}
				}
			}
		}

		public void ProcessPlayers() {
			// This function handles production and processing for each player.

			// produce goods
			if (iTimer % 5 == 0) {
				foreach (ExchangeAgent a in agents) {
					if (a.player.iProduction == 0) {
						if ((eEvent != EventsEnum.DustStorm) || (iTimer % 20 == 0)) {  // halve production during dust storm (event 12)
							a.player.iHoldings[0]++;
							if ((eEvent == EventsEnum.Cooperation) || (eEvent == EventsEnum.SolarFlare))  // excess productivity - try doing it again
								a.player.iHoldings[0]++;
						}
					}
					else if (a.player.iHoldings[0] > 0) {
						if ((a.player.iProduction == 1) ||
							(a.player.iProduction == 2) ||
							((a.player.iProduction == 5)  && ((eEvent != EventsEnum.RampantWharks) || (iTimer % 20 == 0))) ||
							(a.player.iProduction == 6) ||
							((a.player.iProduction == 7) && ((eEvent != EventsEnum.Planetquakes) || (iTimer % 20 == 0))) ||
							((a.player.iProduction == 8) && ((eEvent != EventsEnum.DistilleryClosed) || (iTimer % 20 == 0)))) {
							a.player.iHoldings[0]--;
							a.player.iHoldings[a.player.iProduction]++;
							if ((a.player.iHoldings[0] > 0) && 
								((eEvent == EventsEnum.OreVein) && (a.player.iProduction == 1)) ||
								((eEvent == EventsEnum.SludgePool) && (a.player.iProduction == 2)) ||
								((eEvent == EventsEnum.SpiceHarvest) && (a.player.iProduction == 5)) ||
								((eEvent == EventsEnum.SyrineCave) && (a.player.iProduction == 7)) ||
								((eEvent == EventsEnum.MardiGras) && (a.player.iProduction == 8))) {  // excess productivity - try doing it again
								a.player.iHoldings[0]--;
								a.player.iHoldings[a.player.iProduction]++;
							}
						}
						else if ((a.player.iProduction == 3) && (a.player.iHoldings[1] > 0)) {
							a.player.iHoldings[0]--;
							a.player.iHoldings[1]--;
							a.player.iHoldings[3]++;
							if ((eEvent == EventsEnum.Cooperation) && (a.player.iHoldings[0] > 0) && (a.player.iHoldings[1] > 0)) {  // excess productivity - try doing it again
								a.player.iHoldings[0]--;
								a.player.iHoldings[1]--;
								a.player.iHoldings[3]++;
							}
						}
						else if ((a.player.iProduction == 4) && (a.player.iHoldings[2] > 0)) {
							a.player.iHoldings[0]--;
							a.player.iHoldings[2]--;
							a.player.iHoldings[4]++;
							if ((eEvent == EventsEnum.Cooperation) && (a.player.iHoldings[0] > 0) && (a.player.iHoldings[2] > 0)) {  // excess productivity - try doing it again
								a.player.iHoldings[0]--;
								a.player.iHoldings[2]--;
								a.player.iHoldings[4]++;
							}
						}
					}
				}
			}
			// handle orders
			for (int i = 0; i < agents.Count; i++) {
				ExchangeAgent agent = (ExchangeAgent) agents[i];
				agent.Process();
				if (agent.iBuyGood != -1) {
					Buy(i, agent.iBuyGood, 1);
					agent.iBuyGood = -1;
				}
				if (agent.iSellGood != -1) {
					Sell(i, agent.iSellGood, 1);
					agent.iSellGood = -1;
				}
				agent.player.iProduction = agent.iProduction;
			}
		}

		public void SetValues() {
			// This function calculates the value and rank of each player.
			for (int i = 0; i < agents.Count; i++) {
				long lTotalValue = ((ExchangeAgent) agents[i]).player.iCredits;
				for (int j = 0; j < 9; j++)
					lTotalValue += ((ExchangeAgent) agents[i]).player.iHoldings[j] * ((Good) goods[j]).iPrice;
				((ExchangeAgent) agents[i]).player.lTotalValue = lTotalValue;
			}
			// set rankings
			BitArray bRanked = new BitArray(agents.Count, false);
			for (int rank = 1; rank <= agents.Count; rank++) {
				int max = -1;
				for (int i = 0; i < agents.Count; i++) {
					if ((bRanked[i] == false) && ((max == -1) || (((ExchangeAgent) agents[i]).player.lTotalValue > ((ExchangeAgent) agents[max]).player.lTotalValue)))
						max = i;
				}
				((ExchangeAgent) agents[max]).player.iRank = rank;
				bRanked[max] = true;
			}
		}

		public void UpdateTimer() {
			// This function updates the timer and stops the game when it ends.
			if ((++iTimer) == 60) {
				iTimer = 0;
				if ((++iPhase) == 11) {  // game over
					bRunning = false;
					bDone = true;
					// update agent statistics
					foreach (ExchangeAgent agent in agents) {
						agent.iGames++;
						agent.dRank += agent.player.iRank;
						agent.lValue += agent.player.lTotalValue;
					}
				}
				else
					strNews += " * Phase " + iPhase + " * ";
			}
		}

		#endregion

		#region Helper functions

		public void AddAgent(ExchangeAgent agent) {
			// This function adds an agent to the agents list and creates a player for it.
			agent.player = new Player(agent.strName);
			agent.exchange = this;
			agents.Add(agent);
		}

		public void Buy(int player, int good, int quantity) {
			// This function attempts to purchase the specified quantity of goods.
			// calculate the minimum of (a) the total units the player can afford, (b) the total number of units available, and (c) the player's requested quantity
			int purchase = Math.Min(Math.Min(((ExchangeAgent) agents[player]).player.iCredits / ((Good) goods[good]).iPrice, ((Good) goods[good]).iQuantity), quantity);
			if (purchase > 0) {
				((ExchangeAgent) agents[player]).player.iCredits -= ((Good) goods[good]).iPrice * purchase;
				((Good) goods[good]).iQuantity -= purchase;
				((ExchangeAgent) agents[player]).player.iHoldings[good] += purchase;
			}
		}

		public void Sell(int player, int good, int quantity) {
			// This function attempts to sell the specified quantity of goods.
			int sell = Math.Min(((ExchangeAgent) agents[player]).player.iHoldings[good], quantity);
			if (sell > 0) {
				((ExchangeAgent) agents[player]).player.iCredits += ((Good) goods[good]).iPrice * sell;
				((Good) goods[good]).iQuantity += sell;
				((ExchangeAgent) agents[player]).player.iHoldings[good] -= sell;
			}
		}

		public string Time() {
			// This function returns the current game time as a string.
			return iPhase + ":" + (iTimer / 10) + ":" + (iTimer % 10);
		}

		#endregion

	}

	public abstract class Good {
		// This abstract base class represents a traded good. It is subclassed by each good with a specific SetSpecificPrice function.

		#region Variables and initialization code

		public ExchangeSimulation exchange;			// the exchange on which the good is being traded
		public string strName;													// the name of the good
		public int iQuantity;															// the quantity of the good on the market
		public int iBasePrice;														// the base price of the good (before being modified by current events)
		public int iPrice;																	// the actual price of the good
		public int iTrend;																// the price trend of the good
		public int[] iPriceHistory;												// a record of the price history of the good
		public enum GoodEnum { Power = 0, Ore, Sludge, Galactite, Hydradine, Spice, Shells, Crystals, Rum };

		public Good(ExchangeSimulation exchange, int price, int quantity) {
			this.exchange = exchange;
			this.iPrice = this.iBasePrice = price;
			this.iQuantity = quantity;
			iPriceHistory = new int[100];
			for (int i = 0; i < iPriceHistory.Length; i++)
				iPriceHistory[i] = -10;
		}

		#endregion

		#region Core functions

		public void SetPrice(bool record) {
			// This function sets the price of the good.
			if (record) {  // remember whether or not this price should be recorded
				for (int i = iPriceHistory.Length - 1; i > 0; i--)
					iPriceHistory[i] = iPriceHistory[i - 1];
				iPriceHistory[0] = iPrice;
			}
			iBasePrice = SpecificSetPrice();
			iPrice = iBasePrice;
			// adjust price based on events
			if (exchange.eEvent == ExchangeSimulation.EventsEnum.MerchantConvoy)
				iPrice *= 2;
			else if (exchange.eEvent == ExchangeSimulation.EventsEnum.GoodEconomy)
				iPrice /= 2;
			else if ((exchange.eEvent == ExchangeSimulation.EventsEnum.WharkBreeding) && (this is Good_Shells))
				iPrice = (int) (iPrice * 1.5);
			else if ((exchange.eEvent == ExchangeSimulation.EventsEnum.SaveTheWharks) && (this is Good_Shells))
				iPrice = (int) (iPrice * 0.7);
			// measure and record trend
			iTrend = (iPrice - iPriceHistory[5]);
		}

		public abstract int SpecificSetPrice();			// abstract class overridden for each specific good

		#endregion

	}

	public class Good_Power : Good {
		// This class represents the Solar Power good.

		#region Initialization code

		public Good_Power(ExchangeSimulation exchange, int price, int quantity) : base(exchange, price, quantity) {
			strName = "Power";
		}

		#endregion

		#region Core functions

		public override int SpecificSetPrice() {
			// Power is a commodity good. It fluctuates between 20 and 30 credits unless market supply is low, in which case it steadily rises.
			if (iQuantity < 10)
				return Math.Min(iBasePrice + exchange.random.Next(2), 120 + exchange.random.Next(10));
			else if (iBasePrice > 30)
				return iBasePrice - 2 - exchange.random.Next(2);
			else
				return Math.Max(Math.Min(iBasePrice + exchange.random.Next(3) - 1, 30), 20);
		}

		#endregion

	}

	public class Good_Ore : Good {
		// This class represents the Ore good.

		#region Initialization code

		public Good_Ore(ExchangeSimulation exchange, int price, int quantity) : base(exchange, price, quantity) {
			strName = "Ore";
		}

		#endregion

		#region Core functions

		public override int SpecificSetPrice() {
			// Ore is a commodity good. It fluctuates between 30 and 40 credits unless market supply is low, in which case it steadily rises.
			if (iQuantity < 10)
				return Math.Min(iBasePrice + exchange.random.Next(3), 90 + exchange.random.Next(10));
			else if (iBasePrice > 40)
				return iBasePrice - 2 - exchange.random.Next(2);
			else
				return Math.Max(Math.Min(iBasePrice + exchange.random.Next(3) - 1, 40), 30);
		}

		#endregion

	}

	public class Good_Sludge : Good {
		// This class represents the Sludge good.

		#region Initialization code

		public Good_Sludge(ExchangeSimulation exchange, int price, int quantity) : base(exchange, price, quantity) {
			strName = "Sludge";
		}

		#endregion

		#region Core functions

		public override int SpecificSetPrice() {
			// Sludge is a commodity good. It fluctuates between 25 and 35 credits unless market supply is low, in which case it steadily rises.
			if (iQuantity < 10)
				return Math.Min(iBasePrice + exchange.random.Next(3), 90 + exchange.random.Next(10));
			else if (iBasePrice > 35)
				return iBasePrice - 2 - exchange.random.Next(2);
			else
				return Math.Max(Math.Min(iBasePrice + exchange.random.Next(3) - 1, 35), 25);
		}

		#endregion

	}

	public class Good_Galactite : Good {
		// This class represents the Galactite good.

		#region Variables and initialization code

		private int iCycleDuration, iCycleCurrent, iMagnitude;

		public Good_Galactite(ExchangeSimulation exchange, int price, int quantity) : base(exchange, price, quantity) {
			strName = "Galactite";
			iCycleDuration = exchange.random.Next(121) + 120;  // cyclic range between 120 and 240 cycles (two to four phases)
			iCycleCurrent = exchange.random.Next(iCycleDuration);
			iMagnitude = exchange.random.Next(40) + 30;
		}

		#endregion

		#region Core functions

		public override int SpecificSetPrice() {
			// Galactite trades on a long cycle. The magnitude of the cycle fluctuates randomly.
			iCycleCurrent++;
			iMagnitude = Math.Max(Math.Min(iMagnitude + exchange.random.Next(3) - 1, 70), 30);
			return (int) (120 + iMagnitude * (Math.Sin(2.0 * Math.PI * (double) ((float) iCycleCurrent / (float) iCycleDuration))));
		}

		#endregion

	}

	public class Good_Hydradine: Good {
		// This class represents the Hydradine good.

		#region Variables and initialization code

		private int iMomentum;
		public Good_Hydradine(ExchangeSimulation exchange, int price, int quantity) : base(exchange, price, quantity) {
			strName = "Hydradine";
		}

		#endregion

		#region Core functions

		public override int SpecificSetPrice() {
			// Hydradine trades at a price determined randomly with momentum.
			if (exchange.random.Next(20) == 1)
				iMomentum = -iMomentum;
			iMomentum = Math.Max(Math.Min(iMomentum + exchange.random.Next(3) - 1, 5), -5);
			return Math.Max(Math.Min(iBasePrice + iMomentum, 160 + exchange.random.Next(10)), 40 + exchange.random.Next(10));
		}

		#endregion

	}

	public class Good_Spice : Good {
		// This class represents the Spice good.

		#region Variables and initialization code

		private int iCycleDuration, iCycleCurrent, iMagnitude;

		public Good_Spice(ExchangeSimulation exchange, int price, int quantity) : base(exchange, price, quantity) {
			strName = "Spice";
			iCycleDuration = exchange.random.Next(121) + 90;  // cyclic range between 90 and 210 cycles
			iCycleCurrent = exchange.random.Next(iCycleDuration);
			iMagnitude = exchange.random.Next(30) + 30;
		}

		#endregion

		#region Core functions

		public override int SpecificSetPrice() {
			// Spice is traded with a price that varies over a long cycle.
			iCycleCurrent++;
			iMagnitude = Math.Max(Math.Min(iMagnitude + exchange.random.Next(3) - 1, 70), 30);
			return (int) (80 + iMagnitude * (Math.Sin(2.0 * Math.PI * (double) ((float) iCycleCurrent / (float) iCycleDuration))));
		}

		#endregion

	}

	public class Good_Shells : Good {
		// This class represents the Whark Shells good.

		#region Variables and initialization code

		private int iMomentum;
		public Good_Shells(ExchangeSimulation exchange, int price, int quantity) : base(exchange, price, quantity) {
			strName = "Whark Shells";
		}

		#endregion

		#region Core functions

		public override int SpecificSetPrice() {
			// Whark shells trade on a price randomized with momentum.
			if (exchange.random.Next(20) == 1)
				iMomentum = -iMomentum;
			iMomentum = Math.Max(Math.Min(iMomentum + exchange.random.Next(5) - 2, 6), -6);
			return Math.Max(Math.Min(iBasePrice + iMomentum, 140), 20);
		}

		#endregion

	}

	public class Good_Crystals : Good {
		// This class represents the Syrine Crystals good.

		#region Initialization code

		public Good_Crystals(ExchangeSimulation exchange, int price, int quantity) : base(exchange, price, quantity) {
			strName = "Syrine Crystals";
		}

		#endregion

		#region Core functions

		public override int SpecificSetPrice() {
			// Syrine crystals are a bubble market. The price gradually and steadily rises, but is prone to random, abrupt, and huge collapses.
			if ((exchange.iTimer % 12 == 0) && (exchange.random.Next(180) + 75 < iBasePrice))
				return Math.Max(20 + exchange.random.Next(15), iBasePrice - 30 - exchange.random.Next(90));
			else
				return iBasePrice + exchange.random.Next(3);
		}

		#endregion

	}

	public class Good_Rum : Good {
		// This class represents the Denebian Rum good.

		#region Initialization code

		public Good_Rum(ExchangeSimulation exchange, int price, int quantity) : base(exchange, price, quantity) {
			strName = "Denebian Rum";
		}

		#endregion

		#region Core functions

		public override int SpecificSetPrice() {
			// Denebian Rum trades at a completely random price.
			if ((iBasePrice > 150) && (exchange.random.Next(20) == 0))
				iBasePrice = 120 + exchange.random.Next(20);
			else if ((iBasePrice < 35) && (exchange.random.Next(20) == 0))
				iBasePrice = 40 + exchange.random.Next(20);
			return Math.Min(Math.Max(iBasePrice + exchange.random.Next(5) - 2, 25), 160);
		}

		#endregion

	}

	public class Player {
		// This class represents an Exchange player.

		#region Variables and initialization code

		public string strName;								// The player's name.
		public int iCredits;											// The number of credits the player currently holds.
		public int[] iHoldings;									// The quantity of each good that the player currently owns.
		public int iProduction;								// The good that the player is currently producing.
		public int iRank;												// The player's current rank.
		public long lTotalValue;							// The calculated total value of the player's credits and holdings.

		public Player(string strName) {
			this.strName = strName;
			iHoldings = new int[9];
			iHoldings[0] = 20;  // start with an initial stock of energy
			iCredits = 1000;
			iProduction = 0;
		}

		#endregion

	}
}
