// The Exchange: Genetically Evolving Market Agents 
// Written by David J. Stein, Esq., 11/22/2004. See "License.html" for terms and conditions of use.
//
// This module contains the code for the Agent Arena form, including the agent designer (compiler.)

#region Using declarations

using System;
using System.IO;
using System.Drawing;
using System.Collections;
using System.ComponentModel;
using System.Windows.Forms;

#endregion

namespace The_Exchange_Namespace {

	public class ExchangeAgentDesigner : System.Windows.Forms.Form	{
		// This class represents the Agent Designer dialog.

		#region  Variables and initialization code

		private ExchangeAgent_Programmed agent;
		private System.Windows.Forms.TextBox txtAssembly;
		private System.Windows.Forms.TextBox txtOpcodes;
		private System.Windows.Forms.TextBox txtStatus;
		private System.Windows.Forms.Label label1;
		private System.Windows.Forms.TextBox txtName;
		private System.ComponentModel.Container components = null;
		private System.Windows.Forms.Button btnCompile;
		private System.Windows.Forms.Button btnSave;
		private System.Windows.Forms.Button btnLoad;
		private System.Windows.Forms.Button btnHelp;

		public ExchangeAgentDesigner()	{
			InitializeComponent();
			// setup initial agent
			agent = new ExchangeAgent_Programmed(new Random(0));
			agent.cProgram[0] = (char) 255;
			for (int i = 1; i < agent.cProgram.Length; i++)
				agent.cProgram[i] = (char) 0;
		}

		#endregion

		#region User interface event handlers

		private void txtName_TextChanged(object sender, System.EventArgs e) {
			// This function changes the name of the agent.
			agent.strName = txtName.Text;
		}

		private void btnSave_Click(object sender, System.EventArgs e) {
			// This function saves an agent to disk.
			SaveFileDialog s = new SaveFileDialog();
			s.Filter = "Exchange Agents (*.exa)|*.exa";
			s.OverwritePrompt = true;
			if (s.ShowDialog() == DialogResult.OK)
				agent.Save(s.FileName);
		}

		private void btnLoad_Click(object sender, System.EventArgs e) {
			// This function loads an agent from disk.
			OpenFileDialog open = new OpenFileDialog();
			open.Filter = "Exchange Agents (*.exa)|*.exa";
			if (open.ShowDialog() == DialogResult.OK) {  // selected an XML file
				ExchangeAgent_Programmed newAgent = ExchangeAgent_Programmed.Load(open.FileName);
				if (newAgent != null) {
					agent = newAgent;
					txtAssembly.Text = agent.strAssembly;
					txtAssembly.Refresh();
					txtOpcodes.Text = "";
					ArrayList list = agent.DisassembleProgram();
					foreach (string operation in list)
						txtOpcodes.Text += operation + "\r\n";
					txtOpcodes.Refresh();
					txtName.Text = agent.strName;
					txtName.Refresh();
				}
			}
		}

		private void btnHelp_Click(object sender, System.EventArgs e) {
			// This function handles clicks for the Help button.
			ExchangeAgentDesignerHelpDialog h = new ExchangeAgentDesignerHelpDialog();
			h.Show();
		}

		private void btnCompile_Click(object sender, System.EventArgs e) {
			// This function attempts to compile the assembly code.
			ArrayList opcodes = new ArrayList();
			Hashtable labels = new Hashtable();
			ArrayList references_labels = new ArrayList();
			ArrayList references_addresses = new ArrayList();
			string[] split = txtAssembly.Text.Split(new char[] { '\n' } );
			string error = "";
			// first pass: assemble opcodes, record labels, and record references
			int iProgramCounter = 0;
			foreach (string fullstring in split) {
				string[] splittokens = fullstring.Split(new char[] { ' ', '\t' } );
				// trim tokens and get rid of zero-length ones
				ArrayList tokens = new ArrayList();
				foreach (string token in splittokens) {
					string trimmed = token.Trim();
					if (trimmed.StartsWith("'"))
						break;  // rest of line is a comment
					else if (trimmed.Length > 0)
						tokens.Add(trimmed);
				}
				// now parse
				if (tokens.Count > 0) {
					if (tokens[0].ToString().EndsWith(":"))  {   // this is a label - add to labels list
						string label = tokens[0].ToString().Substring(0, tokens[0].ToString().Length - 1);  // remove end colon
						if (labels.ContainsKey(label) )
							error += "Syntax Error: Line " + iProgramCounter + ": Label " + tokens[0].ToString().Substring(0, tokens[0].ToString().Length - 1) + " already defined\r\n";
						else
							labels.Add(tokens[0].ToString().Substring(0, tokens[0].ToString().Length - 1), iProgramCounter);
						tokens.RemoveAt(0);  // throw away label in tokens list
					}
				}
				if (tokens.Count > 0) {
					if (tokens[0].ToString() == "NOOP") {  // noop
						opcodes.Add((char) 0);
						iProgramCounter++;
					}
					else if (tokens[0].ToString() == "CLR") {  // clear register
						if (tokens.Count != 2)
							error += "Syntax Error: Line " + iProgramCounter + ": CLR improperly formatted\r\n";
						else {
							int register = ParseRegister(tokens[1].ToString());
							if (register ==-1)
								error += "Syntax Error: Line " + iProgramCounter + ": CLR does not specify register\r\n";
							else {  // valid instruction - add
								opcodes.Add((char) (register + 1));
								iProgramCounter++;
							}
						}
					}
					else if (tokens[0].ToString() == "SET") {  // set register, either to register or to constant
						if (tokens.Count != 3)
							error += "Syntax Error: Line " + iProgramCounter + ": SET improperly formatted\r\n";
						else {
							int register1 = ParseRegisterComma(tokens[1].ToString());
							if (register1 ==-1)
								error += "Syntax Error: Line " + iProgramCounter + ": SET does not specify first register\r\n";
							else {
								int register2 = ParseRegister(tokens[2].ToString());
								if (register2 != -1) {
									if (register2 == register1)
										error += "Syntax Error: Line " + iProgramCounter + ": SET sets register to same register\r\n";
									else {  // valid instruction - add
										opcodes.Add((char) (9 + register1 * 3 + (register2 > register1 ? register2 - 1 : register2)));
										iProgramCounter++;
									}
								}
								else {
									int constant = ParseSignedConstant(tokens[2].ToString());
									if (constant == -128)
										error += "Syntax Error: Line " + iProgramCounter + ": SET does not specify register or constant\r\n";
									else {  // valid instruction - add
										opcodes.Add((char) (5 + register1));
										opcodes.Add(constant < 0 ? (char) (256 + constant) : (char) constant);
										iProgramCounter += 2;
									}
								}
							}
						}
					}
					else if (tokens[0].ToString() == "SWAP") {  // swap two registers
						if (tokens.Count != 3)
							error += "Syntax Error: Line " + iProgramCounter + ": SWAP improperly formatted\r\n";
						else {
							int register1 = ParseRegisterComma(tokens[1].ToString());
							if (register1 ==-1)
								error += "Syntax Error: Line " + iProgramCounter + ": SWAP does not specify first register\r\n";
							else {
								int register2 = ParseRegister(tokens[2].ToString());
								if (register2 == -1)
									error += "Syntax Error: Line " + iProgramCounter + ": SWAP does not specify register or constant\r\n"; 
								else if (register2 == register1)
									error += "Syntax Error: Line " + iProgramCounter + ": SWAP swaps register with same register\r\n";
								else {  // valid instruction - add
									opcodes.Add((char) (21 + register1 * 3 + (register2 > register1 ? register2 - 1 : register2)));
									iProgramCounter++;
								}	
							}
						}
					}
					else if (tokens[0].ToString() == "LD") {  // load instruction
						if (tokens.Count != 3)
							error += "Syntax Error: Line " + iProgramCounter + ": LD improperly formatted\r\n";
						else if (tokens[1].ToString() != "R0,")
							error += "Syntax Error: Line " + iProgramCounter + ": LD does not store to R0\r\n";
						else {
							int price = ParsePrice(tokens[2].ToString());
							int trend = ParseTrend(tokens[2].ToString());
							int quantity = ParseQuantity(tokens[2].ToString());
							int holdings = ParseHoldings(tokens[2].ToString());
							if (price != -1) {  // LD R0, PRICEx
								opcodes.Add((char) (price + 33));
								iProgramCounter++;
							}
							else if (trend != -1) {  // LD R0, TRENDx
								opcodes.Add((char) (trend + 42));
								iProgramCounter++;
							}
							else if (quantity != -1) {  // LD R0, QUANTITYx
								opcodes.Add((char) (quantity + 51));
								iProgramCounter++;
							}
							else if (holdings != -1) {  // LD R0, HOLDINGSx
								opcodes.Add((char) (holdings + 60));
								iProgramCounter++;
							}
							else if (tokens[2].ToString() == "CREDITS") {  // LD R0, CREDITS
								opcodes.Add((char) 69);
								iProgramCounter++;
							}
							else if (tokens[2].ToString() == "EVENT") {  // LD R0, EVENT
								opcodes.Add((char) 70);
								iProgramCounter++;
							}
							else if (tokens[2].ToString() == "PHASE") {  // LD R0, PHASE
								opcodes.Add((char) 71);
								iProgramCounter++;
							}
							else if (tokens[2].ToString() == "PLAYERCOUNT") {  // LD R0, PLAYERCOUNT
								opcodes.Add((char) 72);
								iProgramCounter++;
							}
							else if (tokens[2].ToString() == "RANDOM") {  // LD R0, RANDOM
								opcodes.Add((char) 73);
								iProgramCounter++;
							}
							else
								error += "Syntax Error: Line " + iProgramCounter + ": LD with unknown symbol " + tokens[2] + "\r\n";
						}
					}
					else if (tokens[0].ToString() == "ADD") {  // add to register, either from register or from constant
						if (tokens.Count != 3)
							error += "Syntax Error: Line " + iProgramCounter + ": ADD improperly formatted\r\n";
						else {
							int register1 = ParseRegisterComma(tokens[1].ToString());
							if (register1 ==-1)
								error += "Syntax Error: Line " + iProgramCounter + ": ADD does not specify first register\r\n";
							else {
								int register2 = ParseRegister(tokens[2].ToString());
								if (register2 != -1) {
									if (register2 == register1)
										error += "Syntax Error: Line " + iProgramCounter + ": ADD adds register to same register\r\n";
									else {  // valid instruction - add
										opcodes.Add((char) (78 + register1 * 3 + (register2 > register1 ? register2 - 1 : register2)));
										iProgramCounter++;
									}
								}
								else { 
									int constant = ParseConstant(tokens[2].ToString());
									if (constant == -1)
										error += "Syntax Error: Line " + iProgramCounter + ": ADD does not specify register or constant\r\n";
									else {  // valid instruction - add
										opcodes.Add((char) (74 + register1));
										opcodes.Add((char) constant);
										iProgramCounter += 2;
									}
								}
							}
						}
					}
					else if (tokens[0].ToString() == "SUB") {  // subtract from register, either from register or from constant
						if (tokens.Count != 3)
							error += "Syntax Error: Line " + iProgramCounter + ": SUB improperly formatted\r\n";
						else {
							int register1 = ParseRegisterComma(tokens[1].ToString());
							if (register1 ==-1)
								error += "Syntax Error: Line " + iProgramCounter + ": SUB does not specify first register\r\n";
							else {
								int register2 = ParseRegister(tokens[2].ToString());
								if (register2 != -1) {
									if (register2 == register1)
										error += "Syntax Error: Line " + iProgramCounter + ": SUB subtracts register from same register\r\n";
									else {  // valid instruction - add
										opcodes.Add((char) (94 + register1 * 3 + (register2 > register1 ? register2 - 1 : register2)));
										iProgramCounter++;
									}
								}
								else { 
									int constant = ParseConstant(tokens[2].ToString());
									if (constant == -1)
										error += "Syntax Error: Line " + iProgramCounter + ": SUB does not specify register or constant\r\n";
									else {  // valid instruction - add
										opcodes.Add((char) (90 + register1));
										opcodes.Add((char) constant);
										iProgramCounter += 2;
									}
								}
							}
						}
					}
					else if (tokens[0].ToString() == "MLT") {  // multiply register, either by register or by constant
						if (tokens.Count != 3)
							error += "Syntax Error: Line " + iProgramCounter + ": MLT improperly formatted\r\n";
						else {
							int register1 = ParseRegisterComma(tokens[1].ToString());
							if (register1 ==-1)
								error += "Syntax Error: Line " + iProgramCounter + ": MLT does not specify first register\r\n";
							else {
								int register2 = ParseRegister(tokens[2].ToString());
								if (register2 != -1) {
									if (register2 == register1)
										error += "Syntax Error: Line " + iProgramCounter + ": MLT multiples register by same register\r\n";
									else {  // valid instruction - add
										opcodes.Add((char) (110 + register1 * 3 + (register2 > register1 ? register2 - 1 : register2)));
										iProgramCounter++;
									}
								}
								else { 
									int constant = ParseConstant(tokens[2].ToString());
									if (constant == -1)
										error += "Syntax Error: Line " + iProgramCounter + ": MLT does not specify register or constant\r\n";
									else {  // valid instruction - add
										opcodes.Add((char) (106 + register1));
										opcodes.Add((char) constant);
										iProgramCounter += 2;
									}
								}
							}
						}
					}
					else if (tokens[0].ToString() == "DIV") {  // divide register, either by register or by constant
						if (tokens.Count != 3)
							error += "Syntax Error: Line " + iProgramCounter + ": DIV improperly formatted\r\n";
						else {
							int register1 = ParseRegisterComma(tokens[1].ToString());
							if (register1 ==-1)
								error += "Syntax Error: Line " + iProgramCounter + ": DIV does not specify first register\r\n";
							else {
								int register2 = ParseRegister(tokens[2].ToString());
								if (register2 != -1) {
									if (register2 == register1)
										error += "Syntax Error: Line " + iProgramCounter + ": DIV divides register by same register\r\n";
									else {  // valid instruction - add
										opcodes.Add((char) (126 + register1 * 3 + (register2 > register1 ? register2 - 1 : register2)));
										iProgramCounter++;
									}
								}
								else { 
									int constant = ParseConstant(tokens[2].ToString());
									if (constant == -1)
										error += "Syntax Error: Line " + iProgramCounter + ": DIV does not specify register or constant\r\n";
									else if (constant == 0)
										error += "Math Error: Line " + iProgramCounter + ": DIV explicitly divides by zero\r\n";
									else {  // valid instruction - add
										opcodes.Add((char) (122 + register1));
										opcodes.Add((char) constant);
										iProgramCounter += 2;
									}
								}
							}
						}
					}
					else if (tokens[0].ToString() == "JMP") {  // unconditional jump
						if (tokens.Count != 2)
							error += "Syntax Error: Line " + iProgramCounter + ": JMP improperly formatted\r\n";
						else {
							int constant = ParseSignedConstant(tokens[1].ToString());
							if (constant == -128) {  // this is a reference - add as label.
								references_labels.Add(tokens[1].ToString());
								references_addresses.Add((int) (iProgramCounter + 1));
								opcodes.Add((char) 138);
								opcodes.Add((char) 0);  // placeholder - will be filled in later
								iProgramCounter += 2;
							}
							else if ((iProgramCounter + constant >= ExchangeAgent_Programmed.iProgramSize) || (iProgramCounter + constant < 0))
								error += "Memory Error: Line " + iProgramCounter + ": JMP to location outside of memory\r\n";
							else {
								opcodes.Add((char) 138);
								opcodes.Add(constant < 0 ? (char) (256 + constant) : (char) constant);
								iProgramCounter += 2;
							}
						}
					}
					else if (tokens[0].ToString() == "JZ") {  // jump if zero
						if (tokens.Count != 3)
							error += "Syntax Error: Line " + iProgramCounter + ": JZ improperly formatted\r\n";
						else {
							int register = ParseRegisterComma(tokens[1].ToString());
							if (register == -1)
								error += "Syntax Error: Line " + iProgramCounter + ": JZ does not specify first register\r\n";
							else {
								int constant = ParseSignedConstant(tokens[2].ToString());
								if (constant == -128) {  // this is a reference - add as label.
									references_labels.Add(tokens[2].ToString());
									references_addresses.Add((int) (iProgramCounter + 1));
									opcodes.Add((char) (139 + register));
									opcodes.Add((char) 0);  // placeholder - will be filled in later
									iProgramCounter += 2;
								}
								else if ((iProgramCounter + constant >= ExchangeAgent_Programmed.iProgramSize) || (iProgramCounter + constant < 0))
									error += "Memory Error: Line " + iProgramCounter + ": JZ to location outside of memory";
								else {
									opcodes.Add((char) (139 + register));
									opcodes.Add(constant < 0 ? (char) (256 + constant) : (char) constant);
									iProgramCounter += 2;
								}
							}
						}
					}
					else if (tokens[0].ToString() == "JNZ") {  // jump if not zero
						if (tokens.Count != 3)
							error += "Syntax Error: Line " + iProgramCounter + ": JNZ improperly formatted\r\n";
						else {
							int register = ParseRegisterComma(tokens[1].ToString());
							if (register ==-1)
								error += "Syntax Error: Line " + iProgramCounter + ": JNZ does not specify first register\r\n";
							else {
								int constant = ParseSignedConstant(tokens[2].ToString());
								if (constant == -128) {  // this is a reference - add as label.
									references_labels.Add(tokens[2].ToString());
									references_addresses.Add((int) (iProgramCounter + 1));
									opcodes.Add((char) (143 + register));
									opcodes.Add((char) 0);  // placeholder - will be filled in later
									iProgramCounter += 2;
								}
								else if ((iProgramCounter + constant >= ExchangeAgent_Programmed.iProgramSize) || (iProgramCounter + constant < 0))
									error += "Memory Error: Line " + iProgramCounter + ": JNZ to location outside of memory";
								else {
									opcodes.Add((char) (143 + register));
									opcodes.Add(constant < 0 ? (char) (256 + constant) : (char) constant);
									iProgramCounter += 2;
								}
							}
						}
					}
					else if (tokens[0].ToString() == "CMPE") {  // compare register, either to register or to constant
						if (tokens.Count != 3)
							error += "Syntax Error: Line " + iProgramCounter + ": CMPE improperly formatted\r\n";
						else {
							int register1 = ParseRegisterComma(tokens[1].ToString());
							if (register1 ==-1)
								error += "Syntax Error: Line " + iProgramCounter + ": CMPE does not specify first register\r\n";
							else {
								int register2 = ParseRegister(tokens[2].ToString());
								if (register2 != -1) {
									if (register2 == register1)
										error += "Syntax Error: Line " + iProgramCounter + ": CMPE compares register to same register\r\n";
									else {  // valid instruction - add
										opcodes.Add((char) (151 + register1 * 3 + (register2 > register1 ? register2 - 1 : register2)));
										iProgramCounter++;
									}
								}
								else { 
									int constant = ParseConstant(tokens[2].ToString());
									if (constant == -1)
										error += "Syntax Error: Line " + iProgramCounter + ": CMPE does not specify register or constant\r\n";
									else {  // valid instruction - add
										opcodes.Add((char) (147 + register1));
										opcodes.Add((char) constant);
										iProgramCounter += 2;
									}
								}
							}
						}
					}
					else if (tokens[0].ToString() == "CMPNE") {  // compare register, either to register or to constant
						if (tokens.Count != 3)
							error += "Syntax Error: Line " + iProgramCounter + ": CMPNE improperly formatted\r\n";
						else {
							int register1 = ParseRegisterComma(tokens[1].ToString());
							if (register1 ==-1)
								error += "Syntax Error: Line " + iProgramCounter + ": CMPNE does not specify first register\r\n";
							else {
								int register2 = ParseRegister(tokens[2].ToString());
								if (register2 != -1) {
									if (register2 == register1)
										error += "Syntax Error: Line " + iProgramCounter + ": CMPNE compares register to same register\r\n";
									else {  // valid instruction - add
										opcodes.Add((char) (167 + register1 * 3 + (register2 > register1 ? register2 - 1 : register2)));
										iProgramCounter++;
									}
								}
								else { 
									int constant = ParseConstant(tokens[2].ToString());
									if (constant == -1)
										error += "Syntax Error: Line " + iProgramCounter + ": CMPNE does not specify register or constant\r\n";
									else {  // valid instruction - add
										opcodes.Add((char) (163 + register1));
										opcodes.Add((char) constant);
										iProgramCounter += 2;
									}
								}
							}
						}
					}
					else if (tokens[0].ToString() == "CMPG") {  // compare register, either to register or to constant
						if (tokens.Count != 3)
							error += "Syntax Error: Line " + iProgramCounter + ": CMPG improperly formatted\r\n";
						else {
							int register1 = ParseRegisterComma(tokens[1].ToString());
							if (register1 ==-1)
								error += "Syntax Error: Line " + iProgramCounter + ": CMPG does not specify first register\r\n";
							else {
								int register2 = ParseRegister(tokens[2].ToString());
								if (register2 != -1) {
									if (register2 == register1)
										error += "Syntax Error: Line " + iProgramCounter + ": CMPG compares register to same register\r\n";
									else {  // valid instruction - add
										opcodes.Add((char) (183 + register1 * 3 + (register2 > register1 ? register2 - 1 : register2)));
										iProgramCounter++;
									}
								}
								else { 
									int constant = ParseConstant(tokens[2].ToString());
									if (constant == -1)
										error += "Syntax Error: Line " + iProgramCounter + ": CMPG does not specify register or constant\r\n";
									else {  // valid instruction - add
										opcodes.Add((char) (179 + register1));
										opcodes.Add((char) constant);
										iProgramCounter += 2;
									}
								}
							}
						}
					}
					else if (tokens[0].ToString() == "CMPGE") {  // compare register, either to register or to constant
						if (tokens.Count != 3)
							error += "Syntax Error: Line " + iProgramCounter + ": CMPGE improperly formatted\r\n";
						else {
							int register1 = ParseRegisterComma(tokens[1].ToString());
							if (register1 ==-1)
								error += "Syntax Error: Line " + iProgramCounter + ": CMPGE does not specify first register\r\n";
							else {
								int register2 = ParseRegister(tokens[2].ToString());
								if (register2 != -1) {
									if (register2 == register1)
										error += "Syntax Error: Line " + iProgramCounter + ": CMPGE compares register to same register\r\n";
									else {  // valid instruction - add
										opcodes.Add((char) (199 + register1 * 3 + (register2 > register1 ? register2 - 1 : register2)));
										iProgramCounter++;
									}
								}
								else { 
									int constant = ParseConstant(tokens[2].ToString());
									if (constant == -1)
										error += "Syntax Error: Line " + iProgramCounter + ": CMPGE does not specify register or constant\r\n";
									else {  // valid instruction - add
										opcodes.Add((char) (195 + register1));
										opcodes.Add((char) constant);
										iProgramCounter += 2;
									}
								}
							}
						}
					}
					else if (tokens[0].ToString() == "CMPL") {  // compare register, either to register or to constant
						if (tokens.Count != 3)
							error += "Syntax Error: Line " + iProgramCounter + ": CMPL improperly formatted\r\n";
						else {
							int register1 = ParseRegisterComma(tokens[1].ToString());
							if (register1 ==-1)
								error += "Syntax Error: Line " + iProgramCounter + ": CMPL does not specify first register\r\n";
							else {
								int register2 = ParseRegister(tokens[2].ToString());
								if (register2 != -1) {
									if (register2 == register1)
										error += "Syntax Error: Line " + iProgramCounter + ": CMPL compares register to same register\r\n";
									else {  // valid instruction - add
										opcodes.Add((char) (215 + register1 * 3 + (register2 > register1 ? register2 - 1 : register2)));
										iProgramCounter++;
									}
								}
								else { 
									int constant = ParseConstant(tokens[2].ToString());
									if (constant == -1)
										error += "Syntax Error: Line " + iProgramCounter + ": CMPL does not specify register or constant\r\n";
									else {  // valid instruction - add
										opcodes.Add((char) (211 + register1));
										opcodes.Add((char) constant);
										iProgramCounter += 2;
									}
								}
							}
						}
					}
					else if (tokens[0].ToString() == "CMPLE") {  // compare register, either to register or to constant
						if (tokens.Count != 3)
							error += "Syntax Error: Line " + iProgramCounter + ": CMPLE improperly formatted\r\n";
						else {
							int register1 = ParseRegisterComma(tokens[1].ToString());
							if (register1 ==-1)
								error += "Syntax Error: Line " + iProgramCounter + ": CMPLE does not specify first register\r\n";
							else {
								int register2 = ParseRegister(tokens[2].ToString());
								if (register2 != -1) {
									if (register2 == register1)
										error += "Syntax Error: Line " + iProgramCounter + ": CMPLE compares register to same register\r\n";
									else {  // valid instruction - add
										opcodes.Add((char) (231 + register1 * 3 + (register2 > register1 ? register2 - 1 : register2)));
										iProgramCounter++;
									}
								}
								else { 
									int constant = ParseConstant(tokens[2].ToString());
									if (constant == -1)
										error += "Syntax Error: Line " + iProgramCounter + ": CMPLE does not specify register or constant\r\n";
									else {  // valid instruction - add
										opcodes.Add((char) (227 + register1));
										opcodes.Add((char) constant);
										iProgramCounter += 2;
									}
								}
							}
						}
					}
					else if (tokens[0].ToString() == "ST") {  // store register
						if (tokens.Count != 3)
							error += "Syntax Error: Line " + iProgramCounter + ": ST improperly formatted\r\n";
						else {
							int register = ParseRegister(tokens[2].ToString());
							if (register ==-1)
								error += "Syntax Error: Line " + iProgramCounter + ": CLR does not specify register\r\n";
							else {  // valid instruction - add
								if (tokens[1].ToString() == "BUY,") {  // store buy
									opcodes.Add((char) (243 + register));
									iProgramCounter++;
								}
								else if (tokens[1].ToString() == "SELL,") {  // store sell
									opcodes.Add((char) (247 + register));
									iProgramCounter++;
								}
								else if (tokens[1].ToString() == "PRODUCTION,") {  // store production
									opcodes.Add((char) (251 + register));
									iProgramCounter++;
								}
								else
									error += "Syntax Error: Line " + iProgramCounter + ": Unknown destination " + tokens[1].ToString() + "\r\n";
							}
						}
					}
					else if (tokens[0].ToString() == "END") {  // END instruction
						opcodes.Add((char) 255);
						iProgramCounter++;
						break;  // done processing instructions
					}
					else
						error += "Syntax Error: Line " + iProgramCounter + ": Unknown operation " + tokens[0].ToString() + "\r\n";
				}
			}
			// check results
			if (opcodes.Count > ExchangeAgent_Programmed.iProgramSize)
				error += "Memory Error: Program exceeds max size of " + ExchangeAgent_Programmed.iProgramSize + " opcodes\r\n";
			if (error.Length == 0) {  // pass 1 succeeded - try pass 2 (fill in links)
				for (int i = 0; i < opcodes.Count; i++)
					agent.cProgram[i] = (char) opcodes[i];
				for (int i = opcodes.Count; i < ExchangeAgent_Programmed.iProgramSize; i++)
					agent.cProgram[i] = (char) 0;
				for (int i = 0; i < references_labels.Count; i++) {
					if (labels.ContainsKey(references_labels[i].ToString()) == false)
						error += "Syntax Error: Line " + references_addresses[i].ToString() + ": Unknown label " + references_labels[i].ToString() + "\r\n";
					else {
						int iComputedAddress = (int) labels[references_labels[i].ToString()] - (int) references_addresses[i];
						if ((iComputedAddress > 128) || (iComputedAddress < -127))
							error += "Memory Error: Line " + references_addresses[i] + ": Unreachable jump ("  + references_labels[i].ToString() + " is out of range)";
						else
							agent.cProgram[(int) references_addresses[i]] = (char) (iComputedAddress < 0 ? 257 + iComputedAddress : iComputedAddress - 1);
					}
				}
			}
			if (error.Length > 0) {
				txtStatus.Text = "Status: Build failed.";
				txtStatus.Refresh();
				txtOpcodes.Text = error;
				txtOpcodes.Refresh();
			}
			else {  // successful compilation
				txtStatus.Text = "Status: Build succeeded.";
				txtStatus.Refresh();
				txtOpcodes.Text = "";
				ArrayList list = agent.DisassembleProgram();
				foreach (string operation in list)
					txtOpcodes.Text += operation + "\r\n";
				txtOpcodes.Refresh();
				agent.strAssembly = txtAssembly.Text;
			}
		}

		private void ExchangeAgentDesigner_Resize(object sender, System.EventArgs e) {
			// This function handles the Resize event.
			txtAssembly.Width = txtOpcodes.Width = (Width - 48) / 2;
			txtOpcodes.Left = txtAssembly.Location.X + txtAssembly.Width + 16;
		}

		#endregion

		#region Helper functions

		public int ParseRegister(string register) {
			// This function parses a string parameter into the proper register (or -1 if error.)
			if (register == "R0")
				return 0;
			else if (register == "R1")
				return 1;
			else if (register == "R2")
				return 2;
			else if (register == "R3")
				return 3;
			else
				return -1;
		}

		public int ParseRegisterComma(string register) {
			// This function parses a string parameter appended with a comma into the proper register (or -1 if error.)
			if (register == "R0,")
				return 0;
			else if (register == "R1,")
				return 1;
			else if (register == "R2,")
				return 2;
			else if (register == "R3,")
				return 3;
			else
				return -1;
		}

		public int ParseConstant(string register) {
			// This function parses a constant into an unsigned char (or -1 if error.)
			int number = 0;
			for (int i = 0; i < register.Length; i++) {
				if ((register[i] >= '0') && (register[i] <= '9'))
					number = number * 10 + Int32.Parse(register[i].ToString());
				else
					return -1;
			}
			if (number < 256)
				return number;
			else
				return -1;  // overflow
		}

		public int ParseSignedConstant(string register) {
			// This function parses a constant into a signed char (or -128 if error.)
			int number = 0;
			bool negative = false;
			for (int i = 0; i < register.Length; i++) {
				if ((register[i] >= '0') && (register[i] <= '9'))
					number = number * 10 + Int32.Parse(register[i].ToString());
				else if ((i == 0) && (register[i] == '-'))
					negative = true;
				else
					return -128;
			}
			if ((number <= 128) && (number >= -127))
				return (negative ? -number : number);
			else
				return -128;  // overflow
		}

		public int ParsePrice(string register) {
			// This function parses a price code into a good (or -1 if error.)
			if (register == "PRICE1")
				return 0;
			else if (register == "PRICE2")
				return 1;
			else if (register == "PRICE3")
				return 2;
			else if (register == "PRICE4")
				return 3;
			else if (register == "PRICE5")
				return 4;
			else if (register == "PRICE6")
				return 5;
			else if (register == "PRICE7")
				return 6;
			else if (register == "PRICE8")
				return 7;
			else if (register == "PRICE9")
				return 8;
			else
				return -1;
		}

		public int ParseTrend(string register) {
			// This function parses a trend code into a good (or -1 if error.)
			if (register == "TREND1")
				return 0;
			else if (register == "TREND2")
				return 1;
			else if (register == "TREND3")
				return 2;
			else if (register == "TREND4")
				return 3;
			else if (register == "TREND5")
				return 4;
			else if (register == "TREND6")
				return 5;
			else if (register == "TREND7")
				return 6;
			else if (register == "TREND8")
				return 7;
			else if (register == "TREND9")
				return 8;
			else
				return -1;
		}

		public int ParseQuantity(string register) {
			// This function parses a quantity code into a good (or -1 if error.)
			if (register == "QUANTITY1")
				return 0;
			else if (register == "QUANTITY2")
				return 1;
			else if (register == "QUANTITY3")
				return 2;
			else if (register == "QUANTITY4")
				return 3;
			else if (register == "QUANTITY5")
				return 4;
			else if (register == "QUANTITY6")
				return 5;
			else if (register == "QUANTITY7")
				return 6;
			else if (register == "QUANTITY8")
				return 7;
			else if (register == "QUANTITY9")
				return 8;
			else
				return -1;
		}

		public int ParseHoldings(string register) {
			// This function parses a holdings code into a good (or -1 if error.)
			if (register == "HOLDINGS1")
				return 0;
			else if (register == "HOLDINGS2")
				return 1;
			else if (register == "HOLDINGS3")
				return 2;
			else if (register == "HOLDINGS4")
				return 3;
			else if (register == "HOLDINGS5")
				return 4;
			else if (register == "HOLDINGS6")
				return 5;
			else if (register == "HOLDINGS7")
				return 6;
			else if (register == "HOLDINGS8")
				return 7;
			else if (register == "HOLDINGS9")
				return 8;
			else
				return -1;
		}

		#endregion

		#region Uninteresting Windows Form Designer generated code

		protected override void Dispose(bool disposing)		{
			if (disposing) {
				if (components != null) {
					components.Dispose();
				}
			}
			base.Dispose( disposing );
		}

		private void InitializeComponent()	{
			System.Resources.ResourceManager resources = new System.Resources.ResourceManager(typeof(ExchangeAgentDesigner));
			this.txtAssembly = new System.Windows.Forms.TextBox();
			this.txtOpcodes = new System.Windows.Forms.TextBox();
			this.btnCompile = new System.Windows.Forms.Button();
			this.txtStatus = new System.Windows.Forms.TextBox();
			this.label1 = new System.Windows.Forms.Label();
			this.txtName = new System.Windows.Forms.TextBox();
			this.btnSave = new System.Windows.Forms.Button();
			this.btnLoad = new System.Windows.Forms.Button();
			this.btnHelp = new System.Windows.Forms.Button();
			this.SuspendLayout();
			// 
			// txtAssembly
			// 
			this.txtAssembly.Anchor = ((System.Windows.Forms.AnchorStyles)(((System.Windows.Forms.AnchorStyles.Top | System.Windows.Forms.AnchorStyles.Bottom) 
				| System.Windows.Forms.AnchorStyles.Left)));
			this.txtAssembly.Location = new System.Drawing.Point(16, 16);
			this.txtAssembly.Multiline = true;
			this.txtAssembly.Name = "txtAssembly";
			this.txtAssembly.ScrollBars = System.Windows.Forms.ScrollBars.Vertical;
			this.txtAssembly.Size = new System.Drawing.Size(352, 408);
			this.txtAssembly.TabIndex = 0;
			this.txtAssembly.Text = "";
			// 
			// txtOpcodes
			// 
			this.txtOpcodes.Anchor = ((System.Windows.Forms.AnchorStyles)(((System.Windows.Forms.AnchorStyles.Top | System.Windows.Forms.AnchorStyles.Bottom) 
				| System.Windows.Forms.AnchorStyles.Left)));
			this.txtOpcodes.Location = new System.Drawing.Point(384, 16);
			this.txtOpcodes.Multiline = true;
			this.txtOpcodes.Name = "txtOpcodes";
			this.txtOpcodes.ReadOnly = true;
			this.txtOpcodes.ScrollBars = System.Windows.Forms.ScrollBars.Vertical;
			this.txtOpcodes.Size = new System.Drawing.Size(352, 408);
			this.txtOpcodes.TabIndex = 5;
			this.txtOpcodes.Text = "";
			// 
			// btnCompile
			// 
			this.btnCompile.Anchor = ((System.Windows.Forms.AnchorStyles)((System.Windows.Forms.AnchorStyles.Bottom | System.Windows.Forms.AnchorStyles.Left)));
			this.btnCompile.Location = new System.Drawing.Point(384, 440);
			this.btnCompile.Name = "btnCompile";
			this.btnCompile.Size = new System.Drawing.Size(72, 23);
			this.btnCompile.TabIndex = 6;
			this.btnCompile.Text = "Compile";
			this.btnCompile.Click += new System.EventHandler(this.btnCompile_Click);
			// 
			// txtStatus
			// 
			this.txtStatus.Anchor = ((System.Windows.Forms.AnchorStyles)(((System.Windows.Forms.AnchorStyles.Bottom | System.Windows.Forms.AnchorStyles.Left) 
				| System.Windows.Forms.AnchorStyles.Right)));
			this.txtStatus.Location = new System.Drawing.Point(464, 440);
			this.txtStatus.Name = "txtStatus";
			this.txtStatus.ReadOnly = true;
			this.txtStatus.Size = new System.Drawing.Size(192, 20);
			this.txtStatus.TabIndex = 7;
			this.txtStatus.Text = "Status: Ready.";
			// 
			// label1
			// 
			this.label1.Anchor = ((System.Windows.Forms.AnchorStyles)((System.Windows.Forms.AnchorStyles.Bottom | System.Windows.Forms.AnchorStyles.Left)));
			this.label1.Location = new System.Drawing.Point(16, 440);
			this.label1.Name = "label1";
			this.label1.Size = new System.Drawing.Size(40, 23);
			this.label1.TabIndex = 1;
			this.label1.Text = "Name";
			this.label1.TextAlign = System.Drawing.ContentAlignment.MiddleLeft;
			// 
			// txtName
			// 
			this.txtName.Anchor = ((System.Windows.Forms.AnchorStyles)((System.Windows.Forms.AnchorStyles.Bottom | System.Windows.Forms.AnchorStyles.Left)));
			this.txtName.Location = new System.Drawing.Point(64, 440);
			this.txtName.Name = "txtName";
			this.txtName.Size = new System.Drawing.Size(136, 20);
			this.txtName.TabIndex = 2;
			this.txtName.Text = "";
			this.txtName.TextChanged += new System.EventHandler(this.txtName_TextChanged);
			// 
			// btnSave
			// 
			this.btnSave.Anchor = ((System.Windows.Forms.AnchorStyles)((System.Windows.Forms.AnchorStyles.Bottom | System.Windows.Forms.AnchorStyles.Left)));
			this.btnSave.Location = new System.Drawing.Point(216, 440);
			this.btnSave.Name = "btnSave";
			this.btnSave.TabIndex = 3;
			this.btnSave.Text = "Save";
			this.btnSave.Click += new System.EventHandler(this.btnSave_Click);
			// 
			// btnLoad
			// 
			this.btnLoad.Anchor = ((System.Windows.Forms.AnchorStyles)((System.Windows.Forms.AnchorStyles.Bottom | System.Windows.Forms.AnchorStyles.Left)));
			this.btnLoad.Location = new System.Drawing.Point(296, 440);
			this.btnLoad.Name = "btnLoad";
			this.btnLoad.TabIndex = 4;
			this.btnLoad.Text = "Load";
			this.btnLoad.Click += new System.EventHandler(this.btnLoad_Click);
			// 
			// btnHelp
			// 
			this.btnHelp.Anchor = ((System.Windows.Forms.AnchorStyles)((System.Windows.Forms.AnchorStyles.Bottom | System.Windows.Forms.AnchorStyles.Left)));
			this.btnHelp.Location = new System.Drawing.Point(664, 440);
			this.btnHelp.Name = "btnHelp";
			this.btnHelp.Size = new System.Drawing.Size(72, 23);
			this.btnHelp.TabIndex = 8;
			this.btnHelp.Text = "Help";
			this.btnHelp.Click += new System.EventHandler(this.btnHelp_Click);
			// 
			// ExchangeAgentDesigner
			// 
			this.AutoScaleBaseSize = new System.Drawing.Size(5, 13);
			this.ClientSize = new System.Drawing.Size(744, 478);
			this.Controls.Add(this.btnHelp);
			this.Controls.Add(this.btnLoad);
			this.Controls.Add(this.btnSave);
			this.Controls.Add(this.txtName);
			this.Controls.Add(this.label1);
			this.Controls.Add(this.txtStatus);
			this.Controls.Add(this.btnCompile);
			this.Controls.Add(this.txtOpcodes);
			this.Controls.Add(this.txtAssembly);
			this.Icon = ((System.Drawing.Icon)(resources.GetObject("$this.Icon")));
			this.Name = "ExchangeAgentDesigner";
			this.StartPosition = System.Windows.Forms.FormStartPosition.CenterScreen;
			this.Text = "Exchange Agent Designer";
			this.Resize += new System.EventHandler(this.ExchangeAgentDesigner_Resize);
			this.ResumeLayout(false);

		}
		#endregion

	}
}
